from __future__ import annotations

import multiprocessing
from concurrent.futures import ProcessPoolExecutor

import pytest

from academy.agent import action
from academy.agent import Agent
from academy.exchange.cloud import spawn_http_exchange
from academy.handle import Handle
from academy.manager import Manager
from academy.socket import open_port


class Coordinator(Agent):
    def __init__(
        self,
        lowerer: Handle[Lowerer],
        reverser: Handle[Reverser],
    ) -> None:
        super().__init__()
        self.lowerer = lowerer
        self.reverser = reverser

    @action
    async def process(self, text: str) -> str:
        text = await self.lowerer.lower(text)
        text = await self.reverser.reverse(text)
        return text


class Lowerer(Agent):
    @action
    async def lower(self, text: str) -> str:
        return text.lower()


class Reverser(Agent):
    @action
    async def reverse(self, text: str) -> str:
        return text[::-1]


@pytest.mark.asyncio
async def test_run_in_processes() -> None:
    with spawn_http_exchange('localhost', open_port()) as factory:
        mp_context = multiprocessing.get_context('spawn')
        executor = ProcessPoolExecutor(max_workers=3, mp_context=mp_context)

        async with await Manager.from_exchange_factory(
            factory=factory,
            executors=executor,
        ) as manager:
            lowerer = await manager.launch(Lowerer)
            reverser = await manager.launch(Reverser)
            coordinator = await manager.launch(
                Coordinator,
                args=(lowerer, reverser),
            )

            text = 'DEADBEEF'
            expected = 'feebdaed'

            result = await coordinator.process(text)
            assert result == expected



from __future__ import annotations

import asyncio
import dataclasses
from concurrent.futures import ThreadPoolExecutor
from typing import Any
from typing import Generic
from typing import TYPE_CHECKING
from typing import TypeVar

from academy.exchange import AgentExchangeClient
from academy.handle import Handle
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.identifier import UserId

if TYPE_CHECKING:
    from academy.agent import AgentT
else:
    AgentT = TypeVar('AgentT')


class ActionContext:
    """Action invocation context."""

    def __init__(
        self,
        source_id: EntityId,
        exchange_client: AgentExchangeClient[Any, Any],
    ) -> None:
        self._source_id = source_id
        self._exchange_client = exchange_client
        self._source_handle: Handle[Any] | None = None

    @property
    def source_id(self) -> EntityId:
        """ID of the source that requested the action."""
        return self._source_id

    @property
    def source_handle(self) -> Handle[Any]:
        """Get a handle to the source agent of the request.

        Returns:
            Handle to the agent that made the request.

        Raises:
            TypeError: If the source is a user entity.
        """
        if isinstance(self.source_id, UserId):
            raise TypeError(
                'Cannot create handle to source because it is a user entity.',
            )
        if self._source_handle is None:
            self._source_handle = Handle(self.source_id)
        return self._source_handle

    def is_agent_source(self) -> bool:
        """Is the source an agent."""
        return isinstance(self.source_id, AgentId)

    def is_user_source(self) -> bool:
        """Is the source a user."""
        return isinstance(self.source_id, UserId)


@dataclasses.dataclass(frozen=True)
class AgentContext(Generic[AgentT]):
    """Agent runtime context."""

    agent_id: AgentId[AgentT]
    """ID of the exchange as registered with the exchange."""
    exchange_client: AgentExchangeClient[AgentT, Any]
    """Client used by agent to communicate with the exchange."""
    executor: ThreadPoolExecutor
    """Thread-pool executor used for running synchronous tasks."""
    shutdown_event: asyncio.Event
    """Shutdown event used to signal the agent to shutdown."""



# ruff: noqa: D102
from __future__ import annotations

import asyncio
import dataclasses
import logging
import sys
from typing import Any
from typing import Generic

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

from aiologic import Lock
from culsans import AsyncQueue
from culsans import AsyncQueueEmpty as QueueEmpty
from culsans import AsyncQueueShutDown as QueueShutDown
from culsans import Queue

from academy.agent import Agent
from academy.agent import AgentT
from academy.exception import BadEntityIdError
from academy.exception import MailboxTerminatedError
from academy.exchange.factory import ExchangeFactory
from academy.exchange.transport import _respond_pending_requests_on_terminate
from academy.exchange.transport import ExchangeTransportMixin
from academy.exchange.transport import MailboxStatus
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.identifier import UserId
from academy.message import Message
from academy.serialize import NoPickleMixin

logger = logging.getLogger(__name__)


class _LocalExchangeState(NoPickleMixin):
    """Local process message exchange.

    LocalExchange is a special case of an exchange where the mailboxes
    of the exchange live in process memory. This class stores the state
    of the exchange.
    """

    def __init__(self) -> None:
        self.queues: dict[EntityId, AsyncQueue[Message[Any]]] = {}
        self.locks: dict[EntityId, Lock] = {}
        self.agents: dict[AgentId[Any], type[Agent]] = {}


@dataclasses.dataclass
class LocalAgentRegistration(Generic[AgentT]):
    """Agent registration for thread exchanges."""

    agent_id: AgentId[AgentT]
    """Unique identifier for the agent created by the exchange."""


class LocalExchangeTransport(ExchangeTransportMixin, NoPickleMixin):
    """Local exchange client bound to a specific mailbox."""

    def __init__(
        self,
        mailbox_id: EntityId,
        state: _LocalExchangeState,
    ) -> None:
        self._mailbox_id = mailbox_id
        self._state = state

    @classmethod
    def new(
        cls,
        mailbox_id: EntityId | None = None,
        *,
        name: str | None = None,
        state: _LocalExchangeState,
    ) -> Self:
        """Instantiate a new transport.

        Args:
            mailbox_id: Bind the transport to the specific mailbox. If `None`,
                a new user entity will be registered and the transport will be
                bound to that mailbox.
            name: Display name of the redistered entity if `mailbox_id` is
                `None`.
            state: Shared state among exchange clients.

        Returns:
            An instantiated transport bound to a specific mailbox.
        """
        if mailbox_id is None:
            mailbox_id = UserId.new(name=name)
            state.queues[mailbox_id] = Queue().async_q
            state.locks[mailbox_id] = Lock()
            logger.info('Registered %s in exchange', mailbox_id)
        return cls(mailbox_id, state)

    @property
    def mailbox_id(self) -> EntityId:
        return self._mailbox_id

    async def close(self) -> None:
        pass

    async def discover(
        self,
        agent: type[Agent],
        *,
        allow_subclasses: bool = True,
    ) -> tuple[AgentId[Any], ...]:
        found: list[AgentId[Any]] = []
        for aid, type_ in self._state.agents.items():
            if agent is type_ or (
                allow_subclasses and issubclass(type_, agent)
            ):
                found.append(aid)
        alive = tuple(
            aid for aid in found if not self._state.queues[aid].is_shutdown
        )
        return alive

    def factory(self) -> LocalExchangeFactory:
        return LocalExchangeFactory(_state=self._state)

    async def recv(self, timeout: float | None = None) -> Message[Any]:
        queue = self._state.queues[self.mailbox_id]
        try:
            return await asyncio.wait_for(queue.get(), timeout=timeout)
        except asyncio.TimeoutError:
            raise TimeoutError(
                f'Timeout waiting for next message for {self.mailbox_id} '
                f'after {timeout} seconds.',
            ) from None
        except QueueShutDown:
            raise MailboxTerminatedError(self.mailbox_id) from None

    async def register_agent(
        self,
        agent: type[AgentT],
        *,
        name: str | None = None,
    ) -> LocalAgentRegistration[AgentT]:
        aid: AgentId[AgentT] = AgentId.new(name=name)
        self._state.queues[aid] = Queue().async_q
        self._state.locks[aid] = Lock()
        self._state.agents[aid] = agent
        return LocalAgentRegistration(agent_id=aid)

    async def send(self, message: Message[Any]) -> None:
        queue = self._state.queues.get(message.dest, None)
        if queue is None:
            raise BadEntityIdError(message.dest)
        async with self._state.locks[message.dest]:
            try:
                await queue.put(message)
            except QueueShutDown:
                raise MailboxTerminatedError(message.dest) from None

    async def status(self, uid: EntityId) -> MailboxStatus:
        if uid not in self._state.queues:
            return MailboxStatus.MISSING
        async with self._state.locks[uid]:
            if self._state.queues[uid].is_shutdown:
                return MailboxStatus.TERMINATED
            return MailboxStatus.ACTIVE

    async def terminate(self, uid: EntityId) -> None:
        queue = self._state.queues.get(uid, None)
        if queue is None:
            return

        async with self._state.locks[uid]:
            if queue.is_shutdown:
                return

            messages = await _drain_queue(queue)
            await _respond_pending_requests_on_terminate(messages, self)

            queue.shutdown(immediate=True)
            if isinstance(uid, AgentId):
                self._state.agents.pop(uid, None)


class LocalExchangeFactory(
    ExchangeFactory[LocalExchangeTransport],
    NoPickleMixin,
):
    """Local exchange client factory.

    A thread exchange can be used to pass messages between agents running
    in separate threads of a single process.
    """

    def __init__(
        self,
        *,
        _state: _LocalExchangeState | None = None,
    ):
        self._state = _LocalExchangeState() if _state is None else _state

    async def _create_transport(
        self,
        mailbox_id: EntityId | None = None,
        *,
        name: str | None = None,
        registration: LocalAgentRegistration[Any] | None = None,  # type: ignore[override]
    ) -> LocalExchangeTransport:
        return LocalExchangeTransport.new(
            mailbox_id,
            name=name,
            state=self._state,
        )


async def _drain_queue(queue: AsyncQueue[Message[Any]]) -> list[Message[Any]]:
    items: list[Message[Any]] = []

    while True:
        try:
            item = queue.get_nowait()
        except (QueueShutDown, QueueEmpty):
            break
        else:
            items.append(item)
            queue.task_done()

    return items



"""Create self-signed SSL certificates for testing.

Warning:
    None of the functions in this module are safe and should only be used
    for creating temporary self-signed certificates for testing.

Note:
    Based on the examples in the
    [Cryptography docs](https://cryptography.io/en/latest/x509/tutorial/#creating-a-self-signed-certificate).
"""

from __future__ import annotations

import datetime
import pathlib
import ssl
from typing import NamedTuple

import pytest
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.x509 import Certificate
from cryptography.x509.oid import NameOID


class SSLContextFixture(NamedTuple):
    """SSL fixture return type."""

    certfile: str
    keyfile: str
    ssl_context: ssl.SSLContext


@pytest.fixture(scope='session')
def ssl_context(tmp_path_factory: pytest.TempPathFactory) -> SSLContextFixture:
    """Create an SSL context from a self-signed certificate."""
    tmp_path = tmp_path_factory.mktemp('ssl-context-fixture')
    certfile = tmp_path / 'cert.pem'
    keyfile = tmp_path / 'key.pem'
    cert, key = create_self_signed_cert()
    write_cert_key_pair(cert, key, certfile, keyfile)

    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    ssl_context.load_cert_chain(certfile, keyfile=keyfile)

    return SSLContextFixture(str(certfile), str(keyfile), ssl_context)


def create_self_signed_cert() -> tuple[Certificate, RSAPrivateKey]:
    """Creates a self-signed certificate and private key pair."""
    key = generate_private_key(public_exponent=65537, key_size=2048)

    # Various details about who we are. For a self-signed certificate the
    # subject and issuer are always the same.
    subject = issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, 'US'),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, 'Illinois'),
            x509.NameAttribute(NameOID.LOCALITY_NAME, 'Chicago'),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, 'Globus Labs'),
            x509.NameAttribute(NameOID.COMMON_NAME, 'labs.globus.com'),
        ],
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.now(datetime.timezone.utc))
        .not_valid_after(
            # Our certificate will be valid for 10 days
            datetime.datetime.now(datetime.timezone.utc)
            + datetime.timedelta(days=10),
        )
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName('localhost')]),
            critical=False,
            # Sign our certificate with our private key
        )
        .sign(key, hashes.SHA256())
    )

    return cert, key


def write_cert_key_pair(
    cert: Certificate,
    key: RSAPrivateKey,
    certfile: str | pathlib.Path,
    keyfile: str | pathlib.Path,
) -> None:
    """Write a self-signed certificate and private key file to disk."""
    encrypted_key = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )
    with open(keyfile, 'wb') as f:
        f.write(encrypted_key)

    with open(certfile, 'wb') as f:
        f.write(cert.public_bytes(serialization.Encoding.PEM))



from __future__ import annotations

import logging
import pickle

import pytest

from academy.exception import AgentTerminatedError
from academy.exception import ExchangeClientNotFoundError
from academy.exchange import LocalExchangeFactory
from academy.exchange import LocalExchangeTransport
from academy.exchange import UserExchangeClient
from academy.exchange.transport import MailboxStatus
from academy.handle import exchange_context
from academy.handle import Handle
from academy.handle import ProxyHandle
from academy.identifier import AgentId
from academy.identifier import UserId
from academy.manager import Manager
from academy.message import ErrorResponse
from academy.message import Message
from academy.message import PingRequest
from academy.message import ShutdownRequest
from testing.agents import CounterAgent
from testing.agents import EmptyAgent
from testing.agents import ErrorAgent
from testing.constant import TEST_SLEEP_INTERVAL


@pytest.mark.asyncio
async def test_proxy_handle_protocol() -> None:
    agent = EmptyAgent()
    handle = ProxyHandle(agent)
    assert str(agent) in str(handle)
    assert repr(agent) in repr(handle)
    assert await handle.ping() >= 0
    await handle.shutdown()


@pytest.mark.asyncio
async def test_proxy_handle_actions() -> None:
    handle = ProxyHandle(CounterAgent())

    # Via Handle.action()
    await handle.action('add', 1)
    count: int = await handle.action('count')
    assert count == 1

    # Via attribute lookup
    await handle.add(1)
    count = await handle.count()
    assert count == 2  # noqa: PLR2004


@pytest.mark.asyncio
async def test_proxy_handle_action_errors() -> None:
    handle = ProxyHandle(ErrorAgent())

    with pytest.raises(RuntimeError, match='This action always fails.'):
        await handle.action('fails')

    with pytest.raises(AttributeError, match='null'):
        await handle.action('null')

    with pytest.raises(AttributeError, match='null'):
        await handle.null()  # type: ignore[attr-defined]

    handle.agent.foo = 1  # type: ignore[attr-defined]
    with pytest.raises(AttributeError, match='not a method'):
        await handle.foo()  # type: ignore[attr-defined]


@pytest.mark.asyncio
async def test_proxy_handle_agent_shutdown_errors() -> None:
    handle = ProxyHandle(EmptyAgent())
    await handle.shutdown()

    with pytest.raises(AgentTerminatedError):
        await handle.action('test')
    with pytest.raises(AgentTerminatedError):
        await handle.ping()
    with pytest.raises(AgentTerminatedError):
        await handle.shutdown()


@pytest.mark.asyncio
async def test_agent_proxy_handle_serialize() -> None:
    agent = EmptyAgent()
    handle = ProxyHandle(agent)
    reconstructed = pickle.loads(pickle.dumps(handle))
    assert isinstance(reconstructed, ProxyHandle)
    assert type(reconstructed.agent) is EmptyAgent
    assert str(reconstructed) == str(handle)
    assert repr(reconstructed) == repr(handle)
    assert reconstructed.agent_id != handle.agent_id


@pytest.mark.asyncio
async def test_agent_handle_serialize(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(EmptyAgent)
    handle = Handle(registration.agent_id)
    reconstructed = pickle.loads(pickle.dumps(handle))
    assert isinstance(reconstructed, Handle)
    assert reconstructed.agent_id == handle.agent_id
    assert str(reconstructed) == str(handle)
    assert repr(reconstructed) == repr(handle)


@pytest.mark.asyncio
async def test_agent_handle_context() -> None:
    # We cannot use the fixture here because the fixture will create context
    factory = LocalExchangeFactory()
    exchange_client = await factory.create_user_client()
    registration = await exchange_client.register_agent(EmptyAgent)
    handle = Handle(registration.agent_id)

    with pytest.raises(ExchangeClientNotFoundError):
        assert handle.exchange is exchange_client

    exchange_context.set(exchange_client)
    assert handle.exchange is exchange_client

    await exchange_client.close()


@pytest.mark.asyncio
async def test_handle_exchange_registration(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    assert len(exchange_client._handles) == 0

    registration = await exchange_client.register_agent(EmptyAgent)
    handle = Handle(registration.agent_id)

    assert handle.exchange is exchange_client
    assert len(exchange_client._handles) == 0

    handle._register_with_exchange(exchange_client)
    assert len(handle._registered_exchanges) == 1
    assert len(exchange_client._handles) == 1

    # Registration is idempotent
    handle._register_with_exchange(exchange_client)
    assert len(handle._registered_exchanges) == 1
    assert len(exchange_client._handles) == 1


@pytest.mark.asyncio
async def test_agent_handle_reuse(
    manager: Manager[LocalExchangeTransport],
) -> None:
    exchange_client = manager.exchange_client
    factory = exchange_client.factory()
    destination = await manager.launch(CounterAgent)

    handle = Handle(destination.agent_id)
    assert handle.exchange is exchange_client, 'Client not inferred.'
    assert handle.handle_id not in exchange_client._handles

    assert await handle.ping() > 0
    assert handle.handle_id in exchange_client._handles
    assert exchange_client in handle._registered_exchanges

    async with await factory.create_user_client() as new_client:
        # Exchange is updated in the agent
        assert handle.exchange is new_client
        assert await handle.ping() > 0
        assert handle.handle_id in new_client._handles
        assert new_client in handle._registered_exchanges

    # Exchange is reset after
    assert handle.exchange is exchange_client

    await exchange_client.close()


@pytest.mark.asyncio
async def test_client_handle_ping_timeout(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(EmptyAgent)
    handle = Handle(registration.agent_id)
    with pytest.raises(TimeoutError):
        await handle.ping(timeout=TEST_SLEEP_INTERVAL)


@pytest.mark.asyncio
async def test_client_handle_log_bad_response(
    manager: Manager[LocalExchangeTransport],
    caplog,
) -> None:
    handle = await manager.launch(EmptyAgent())
    # Should log two messages but not crash:
    #   - User client got an unexpected ping request from agent client
    #   - Agent client got an unexpected ping response (containing an
    #     error produced by user) with no corresponding handle to
    #     send the response to.
    with caplog.at_level(logging.WARNING):
        await handle.exchange.send(
            Message.create(
                src=handle.agent_id,
                dest=handle.exchange.client_id,
                body=PingRequest(),
            ),
        )
        assert await handle.ping() > 0
        await handle.shutdown()

    assert 'no corresponding handle exists' in caplog.text


@pytest.mark.asyncio
async def test_client_handle_shutdown_ignore_already_termiated_error() -> None:
    handle: Handle[EmptyAgent] = Handle(AgentId.new())

    request = Message.create(
        src=UserId.new(),
        dest=handle.agent_id,
        body=ShutdownRequest(),
    )
    handle._shutdown_requests.add(request.tag)
    response = request.create_response(
        ErrorResponse(exception=AgentTerminatedError(handle.agent_id)),
    )
    await handle._process_response(response)


@pytest.mark.asyncio
async def test_client_handle_shutdown_log_error_response(caplog) -> None:
    handle: Handle[EmptyAgent] = Handle(AgentId.new())

    request = Message.create(
        src=AgentId.new(),
        dest=handle.agent_id,
        body=ShutdownRequest(),
    )
    handle._shutdown_requests.add(request.tag)
    response = request.create_response(
        ErrorResponse(exception=AgentTerminatedError(AgentId.new())),
    )

    with caplog.at_level(logging.ERROR):
        await handle._process_response(response)
    assert f'Failure requesting shutdown for {handle.agent_id}' in caplog.text


@pytest.mark.asyncio
async def test_client_handle_actions(
    manager: Manager[LocalExchangeTransport],
) -> None:
    handle = await manager.launch(CounterAgent())
    assert await handle.ping() > 0

    await handle.action('add', 1)
    count: int = await handle.action('count')
    assert count == 1

    await handle.add(1)
    count = await handle.count()
    assert count == 2  # noqa: PLR2004

    await handle.shutdown()


@pytest.mark.parametrize('terminate', (True, False))
@pytest.mark.asyncio
async def test_client_shutdown_termination(
    terminate: bool,
    manager: Manager[LocalExchangeTransport],
) -> None:
    handle = await manager.launch(EmptyAgent())
    await handle.shutdown(terminate=terminate)
    await manager.wait({handle})
    status = await manager.exchange_client.status(handle.agent_id)
    if terminate:
        assert status == MailboxStatus.TERMINATED
    else:
        assert status == MailboxStatus.ACTIVE


@pytest.mark.asyncio
async def test_client_handle_errors(
    manager: Manager[LocalExchangeTransport],
) -> None:
    handle = await manager.launch(ErrorAgent())
    with pytest.raises(
        RuntimeError,
        match='This action always fails.',
    ):
        await handle.fails()

    with pytest.raises(AttributeError, match='null'):
        await handle.action('null')

    await handle.shutdown()


@pytest.mark.asyncio
async def test_handle_default_exchange() -> None:
    factory = LocalExchangeFactory()
    exchange_client = await factory.create_user_client()
    registration = await exchange_client.register_agent(EmptyAgent)
    handle = Handle(registration.agent_id, exchange=exchange_client)

    assert handle.exchange is exchange_client
    assert repr(exchange_client) in repr(handle)

    async with await factory.create_user_client() as new_client:
        assert handle.exchange is new_client

    assert handle.exchange is exchange_client

    await exchange_client.close()


@pytest.mark.asyncio
async def test_handle_ignore_context() -> None:
    factory = LocalExchangeFactory()
    exchange_client = await factory.create_user_client()
    registration = await exchange_client.register_agent(EmptyAgent)
    handle = Handle(
        registration.agent_id,
        exchange=exchange_client,
        ignore_context=True,
    )

    assert handle.exchange is exchange_client
    assert repr(exchange_client) in repr(handle)

    async with await factory.create_user_client():
        assert handle.exchange is exchange_client

    with pytest.raises(pickle.PicklingError):
        pickle.dumps(handle)

    await exchange_client.close()


@pytest.mark.asyncio
async def test_handle_ignore_context_error(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(EmptyAgent)
    with pytest.raises(ValueError, match='no explicit exchange'):
        Handle(registration.agent_id, ignore_context=True)



from __future__ import annotations

import pickle
from typing import Any

import pytest

from academy.identifier import AgentId
from academy.message import ActionRequest
from academy.message import ActionResponse
from academy.message import ErrorResponse
from academy.message import Header
from academy.message import Message
from academy.message import PingRequest
from academy.message import ShutdownRequest
from academy.message import SuccessResponse


@pytest.mark.parametrize(
    'message_body',
    (
        ActionRequest(action='foo', pargs=(b'bar',)),
        PingRequest(),
        ShutdownRequest(),
    ),
)
def test_request_message(message_body: Any) -> None:
    message = Message.create(
        src=AgentId.new(),
        dest=AgentId.new(),
        body=message_body,
    )
    assert isinstance(str(message), str)
    assert isinstance(repr(message), str)
    jsoned = message.model_dump_json()
    recreated: Message[Any] = Message.model_validate_json(jsoned)
    assert message == recreated
    assert hash(message) == hash(recreated)
    assert message != object()
    pickled = message.model_serialize()
    recreated = Message.model_deserialize(pickled)
    assert message == recreated


@pytest.mark.parametrize(
    'message_body',
    (
        ActionResponse(action='foo', result=b'bar'),
        ErrorResponse(exception=Exception()),
        SuccessResponse(),
    ),
)
def test_response_message(message_body: Any) -> None:
    header = Header(src=AgentId.new(), dest=AgentId.new(), kind='response')
    message: Message[Any] = Message(header=header, body=message_body)
    assert isinstance(str(message), str)
    assert isinstance(repr(message), str)
    jsoned = message.model_dump_json()
    recreated: Message[Any] = Message.model_validate_json(jsoned)
    assert message == recreated
    pickled = message.model_serialize()
    recreated = Message.model_deserialize(pickled)
    assert message == recreated


def test_deserialize_bad_type() -> None:
    pickled = pickle.dumps('string')
    with pytest.raises(
        TypeError,
        match='Deserialized message is not of type Message.',
    ):
        Message.model_deserialize(pickled)


def tests_create_response_from_response_error() -> None:
    message = Message.create(
        src=AgentId.new(),
        dest=AgentId.new(),
        body=SuccessResponse(),
    )
    with pytest.raises(
        ValueError,
        match='Cannot create response header from another response',
    ):
        message.create_response(SuccessResponse())


def test_action_request_lazy_deserialize() -> None:
    request = ActionRequest(action='foo', pargs=('bar',), kargs={'foo': 'bar'})

    json = request.model_dump_json()
    reconstructed = ActionRequest.model_validate_json(json)

    assert isinstance(reconstructed, ActionRequest)
    assert isinstance(reconstructed.pargs, str)
    assert isinstance(reconstructed.kargs, str)

    reconstructed.get_args()
    reconstructed.get_kwargs()

    assert isinstance(reconstructed.pargs, tuple)
    assert isinstance(reconstructed.kargs, dict)


def test_action_response_lazy_deserialize() -> None:
    response = ActionResponse(action='foo', result={'foo': 'bar'})

    json = response.model_dump_json()
    reconstructed = ActionResponse.model_validate_json(json)

    assert isinstance(reconstructed, ActionResponse)
    assert isinstance(reconstructed.result, list)

    reconstructed.get_result()

    assert isinstance(reconstructed.result, dict)


def test_error_response_lazy_deserialize() -> None:
    response = ErrorResponse(exception=Exception('Oops!'))

    json = response.model_dump_json()
    reconstructed = ErrorResponse.model_validate_json(json)

    assert isinstance(reconstructed, ErrorResponse)
    assert isinstance(reconstructed.exception, str)

    reconstructed.get_exception()

    assert isinstance(reconstructed.exception, Exception)



from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator
from typing import Any
from unittest import mock

import pytest
import pytest_asyncio

from academy.exception import BadEntityIdError
from academy.exchange import ExchangeFactory
from academy.exchange import MailboxStatus
from academy.exchange import UserExchangeClient
from academy.handle import Handle
from academy.identifier import AgentId
from academy.identifier import UserId
from academy.message import ErrorResponse
from academy.message import Message
from academy.message import PingRequest
from academy.message import Request
from academy.message import SuccessResponse
from testing.agents import EmptyAgent
from testing.constant import TEST_WAIT_TIMEOUT
from testing.fixture import EXCHANGE_FACTORY_TYPES


@pytest_asyncio.fixture(params=EXCHANGE_FACTORY_TYPES)
async def factory(
    request,
    get_factory,
) -> AsyncGenerator[ExchangeFactory[Any]]:
    return get_factory(request.param)


@pytest_asyncio.fixture(params=EXCHANGE_FACTORY_TYPES)
async def client(
    request,
    get_factory,
) -> AsyncGenerator[ExchangeFactory[Any]]:
    factory = get_factory(request.param)
    client = await factory.create_user_client(start_listener=False)
    try:
        yield client
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_create_user_client(factory: ExchangeFactory[Any]) -> None:
    async with await factory.create_user_client(
        start_listener=False,
    ) as client:
        assert isinstance(repr(client), str)
        assert isinstance(str(client), str)


async def _request_handler(_: Any) -> None:  # pragma: no cover
    pass


@pytest.mark.asyncio
async def test_create_agent_client(factory: ExchangeFactory[Any]) -> None:
    async with await factory.create_user_client(
        start_listener=False,
    ) as client:
        registration = await client.register_agent(EmptyAgent)
        async with await factory.create_agent_client(
            registration,
            _request_handler,
        ) as agent_client:
            assert isinstance(repr(agent_client), str)
            assert isinstance(str(agent_client), str)
            await agent_client.close()  # Idempotent check
        await client.close()  # Idempotent check


@pytest.mark.asyncio
async def test_create_agent_client_unregistered(
    factory: ExchangeFactory[Any],
) -> None:
    async with await factory.create_user_client(
        start_listener=False,
    ) as client:
        registration = await client.register_agent(EmptyAgent)
        registration.agent_id = AgentId.new()
        with pytest.raises(BadEntityIdError):
            await factory.create_agent_client(registration, _request_handler)


@pytest.mark.asyncio
async def test_client_discover(client: UserExchangeClient[Any]) -> None:
    registration = await client.register_agent(EmptyAgent)
    assert await client.discover(EmptyAgent) == (registration.agent_id,)


@pytest.mark.asyncio
async def test_client_get_factory(client: UserExchangeClient[Any]) -> None:
    assert isinstance(client.factory(), ExchangeFactory)


@pytest.mark.asyncio
async def test_client_get_status(client: UserExchangeClient[Any]) -> None:
    uid = UserId.new()
    assert await client.status(uid) == MailboxStatus.MISSING
    registration = await client.register_agent(EmptyAgent)
    agent_id = registration.agent_id
    assert await client.status(agent_id) == MailboxStatus.ACTIVE
    await client.terminate(agent_id)
    assert await client.status(agent_id) == MailboxStatus.TERMINATED


@pytest.mark.asyncio
async def test_client_to_agent_message(factory: ExchangeFactory[Any]) -> None:
    received = asyncio.Event()

    async def _handler(_: Message[Request]) -> None:
        received.set()

    async with await factory.create_user_client(
        start_listener=False,
    ) as user_client:
        registration = await user_client.register_agent(EmptyAgent)
        async with await factory.create_agent_client(
            registration,
            _handler,
        ) as agent_client:
            task = asyncio.Task(agent_client._listen_for_messages())

            message = Message.create(
                src=user_client.client_id,
                dest=agent_client.client_id,
                body=PingRequest(),
            )
            await user_client.send(message)

            await asyncio.wait_for(received.wait(), timeout=TEST_WAIT_TIMEOUT)

            await user_client.terminate(registration.agent_id)
            await task


@pytest.mark.asyncio
async def test_agent_handle_process_response(
    factory: ExchangeFactory[Any],
) -> None:
    async def _handler(_: Message[Request]) -> None:  # pragma: no cover
        pass

    async with await factory.create_user_client(
        start_listener=False,
    ) as user_client:
        registration = await user_client.register_agent(EmptyAgent)
        async with await factory.create_agent_client(
            registration,
            _handler,
        ) as agent_client:
            handle: Handle[EmptyAgent] = Handle(AgentId.new())
            handle._register_with_exchange(agent_client)
            message = Message.create(
                src=user_client.client_id,
                dest=agent_client.client_id,
                body=SuccessResponse(),
                label=handle.handle_id,
            )

            with mock.patch.object(handle, '_process_response') as mocked:
                await agent_client._handle_message(message)

            mocked.assert_called()


@pytest.mark.asyncio
async def test_client_reply_error_on_request(
    factory: ExchangeFactory[Any],
) -> None:
    async with await factory.create_user_client(
        start_listener=False,
    ) as client1:
        async with await factory.create_user_client(
            start_listener=True,
        ) as client2:
            message = Message.create(
                src=client1.client_id,
                dest=client2.client_id,
                body=PingRequest(),
            )
            await client1.send(message)
            response = await client1._transport.recv()
            body = response.get_body()
            assert isinstance(body, ErrorResponse)
            assert isinstance(body.get_exception(), TypeError)



from __future__ import annotations

import sys
from collections.abc import Coroutine

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import assert_type
else:  # pragma: <3.11 cover
    from typing_extensions import assert_type

import pytest

from academy.agent import action
from academy.agent import Agent
from academy.handle import Handle
from academy.handle import ProxyHandle
from academy.identifier import AgentId
from academy.mypy_plugin import is_handle_type


@pytest.mark.parametrize(
    ('type_name', 'expected'),
    (
        ('academy.handle', False),
        ('academy.handle.Handle', True),
        ('academy.handle.ProxyHandle', True),
        ('academy.handle.Handle[Foo]', True),
        ('academy.handle.Handle[Foo, Bar]', False),
        ('academy.handle.0Handle', False),
    ),
)
def test_is_handle_type(type_name: str, expected: bool) -> None:
    assert is_handle_type(type_name) == expected


class Example(Agent):
    def __init__(self) -> None:
        self.value = 42

    @action
    async def method(self) -> int:
        return 42

    @action
    async def method_with_arg(self, value: int) -> int:
        return value


@pytest.mark.asyncio
async def test_handle_attribute_access_resolved() -> None:
    handle = ProxyHandle(Example())

    # Attributes/method defined on Handle should stay the same
    assert_type(handle.agent_id, AgentId[Example])
    assert_type(await handle.ping(), float)

    # Methods defined on the Agent should have their return type
    # wrapped in a future
    coro = handle.method()
    assert_type(coro, Coroutine[None, None, int])  # type: ignore[unused-coroutine]
    result = await coro
    assert_type(result, int)


@pytest.mark.asyncio
async def test_handle_attribute_access_error() -> None:
    handle = ProxyHandle(Example())

    # Attribute foo does not exist on Example so should raise an error
    with pytest.raises(AttributeError):
        _ = handle.foo  # type: ignore[attr-defined]

    # Method foo does not exist on Example so should raise an error
    with pytest.raises(AttributeError):
        handle.foo()  # type: ignore[attr-defined]

    # Attribute value does exist on Example but is not accessible via a handle
    with pytest.raises(AttributeError):
        _ = handle.value  # type: ignore[attr-defined]

    # Mypy should catch the bad argument type even if it is not an error
    # at runtime
    await handle.method_with_arg('not-an-int')  # type: ignore[arg-type]


def test_handle_union_types() -> None:
    handle1: Example | ProxyHandle[Example] = Example()
    handle2: ProxyHandle[Example] | Example = Example()

    with pytest.raises(AttributeError):
        handle1.ping()  # type: ignore[union-attr]
    with pytest.raises(AttributeError):
        handle2.ping()  # type: ignore[union-attr]

    handle3: Example | ProxyHandle[Example] = ProxyHandle(Example())
    handle4: ProxyHandle[Example] | Example = ProxyHandle(Example())

    # Case 1: Attribute exists on Example but not ProxyHandle[Example]
    with pytest.raises(AttributeError):
        _ = handle3.value  # type: ignore[union-attr]
    with pytest.raises(AttributeError):
        _ = handle4.value  # type: ignore[union-attr]

    # Case 2: Attribute exists on neither Example nor ProxyHandle[Example]
    with pytest.raises(AttributeError):
        _ = handle3.foo  # type: ignore[union-attr]
    with pytest.raises(AttributeError):
        _ = handle4.foo  # type: ignore[union-attr]


@pytest.mark.asyncio
async def test_handle_generic_protocol() -> None:
    # This is testing the casting a Handle implementation (i.e., ProxyHandle)
    # to the protocol Handle type still preserves the method resolution.
    async def _call(handle: Handle[Example]) -> int:
        ping = await handle.ping()
        assert_type(ping, float)

        result = await handle.method()
        assert_type(result, int)
        return result

    handle = ProxyHandle(Example())
    result = await _call(handle)
    assert_type(result, int)



from __future__ import annotations

import asyncio

import pytest

from academy.event import wait_event_async
from testing.constant import TEST_SLEEP_INTERVAL


@pytest.mark.asyncio
async def test_wait_single_event_set_immediately():
    event = asyncio.Event()
    event.set()
    result = await wait_event_async(event)
    assert result is event


@pytest.mark.asyncio
async def test_wait_single_event_set_later():
    event = asyncio.Event()

    async def set_event_later():
        await asyncio.sleep(TEST_SLEEP_INTERVAL)
        event.set()

    task = asyncio.create_task(set_event_later())
    result = await wait_event_async(event)
    await task
    assert result is event


@pytest.mark.asyncio
async def test_wait_multiple_events_first_one_set():
    event1 = asyncio.Event()
    event2 = asyncio.Event()
    event1.set()
    result = await wait_event_async(event1, event2)
    assert result is event1


@pytest.mark.asyncio
async def test_wait_multiple_events_second_one_set_first():
    event1 = asyncio.Event()
    event2 = asyncio.Event()

    async def set_second():
        await asyncio.sleep(TEST_SLEEP_INTERVAL)
        event2.set()

    task = asyncio.create_task(set_second())
    result = await wait_event_async(event1, event2)
    await task
    assert result is event2


@pytest.mark.asyncio
async def test_wait_timeout_raises():
    event = asyncio.Event()
    with pytest.raises(TimeoutError):
        await wait_event_async(event, timeout=TEST_SLEEP_INTERVAL)



# ruff: noqa: D102
from __future__ import annotations

import dataclasses
import enum
import logging
import sys
import uuid
from typing import Any
from typing import Generic
from typing import NamedTuple

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

import redis.asyncio

from academy.agent import Agent
from academy.agent import AgentT
from academy.exception import BadEntityIdError
from academy.exception import MailboxTerminatedError
from academy.exchange.factory import ExchangeFactory
from academy.exchange.transport import _respond_pending_requests_on_terminate
from academy.exchange.transport import ExchangeTransportMixin
from academy.exchange.transport import MailboxStatus
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.identifier import UserId
from academy.message import Message
from academy.serialize import NoPickleMixin

logger = logging.getLogger(__name__)

_CLOSE_SENTINEL = b'<CLOSED>'


class _RedisConnectionInfo(NamedTuple):
    hostname: str
    port: int
    kwargs: dict[str, Any]


class _MailboxState(enum.Enum):
    ACTIVE = 'ACTIVE'
    INACTIVE = 'INACTIVE'


@dataclasses.dataclass
class RedisAgentRegistration(Generic[AgentT]):
    """Agent registration for redis exchanges."""

    agent_id: AgentId[AgentT]
    """Unique identifier for the agent created by the exchange."""


class RedisExchangeTransport(ExchangeTransportMixin, NoPickleMixin):
    """Redis exchange transport bound to a specific mailbox."""

    def __init__(
        self,
        mailbox_id: EntityId,
        redis_client: redis.asyncio.Redis,
        *,
        redis_info: _RedisConnectionInfo,
    ) -> None:
        self._mailbox_id = mailbox_id
        self._client = redis_client
        self._redis_info = redis_info

    def _active_key(self, uid: EntityId) -> str:
        return f'active:{uid.uid}'

    def _agent_key(self, uid: AgentId[Any]) -> str:
        return f'agent:{uid.uid}'

    def _queue_key(self, uid: EntityId) -> str:
        return f'queue:{uid.uid}'

    @classmethod
    async def new(
        cls,
        *,
        mailbox_id: EntityId | None = None,
        name: str | None = None,
        redis_info: _RedisConnectionInfo,
    ) -> Self:
        """Instantiate a new transport.

        Args:
            mailbox_id: Bind the transport to the specific mailbox. If `None`,
                a new user entity will be registered and the transport will be
                bound to that mailbox.
            name: Display name of the redistered entity if `mailbox_id` is
                `None`.
            redis_info: Redis connection information.

        Returns:
            An instantiated transport bound to a specific mailbox.

        Raises:
            redis.exceptions.ConnectionError: If the Redis server is not
                reachable.
        """
        client = redis.asyncio.Redis(
            host=redis_info.hostname,
            port=redis_info.port,
            decode_responses=False,
            **redis_info.kwargs,
        )
        # Ensure the redis server is reachable else fail early
        await client.ping()

        if mailbox_id is None:
            mailbox_id = UserId.new(name=name)
            await client.set(
                f'active:{mailbox_id.uid}',
                _MailboxState.ACTIVE.value,
            )
            logger.info('Registered %s in exchange', mailbox_id)
        return cls(mailbox_id, client, redis_info=redis_info)

    @property
    def mailbox_id(self) -> EntityId:
        return self._mailbox_id

    async def close(self) -> None:
        await self._client.aclose()

    async def discover(
        self,
        agent: type[Agent],
        *,
        allow_subclasses: bool = True,
    ) -> tuple[AgentId[Any], ...]:
        found: list[AgentId[Any]] = []
        fqp = f'{agent.__module__}.{agent.__name__}'
        async for key in self._client.scan_iter(
            'agent:*',
        ):  # pragma: no branch
            mro_str = await self._client.get(key)
            assert isinstance(mro_str, str)
            mro = mro_str.split(',')
            if fqp == mro[0] or (allow_subclasses and fqp in mro):
                aid: AgentId[Any] = AgentId(uid=uuid.UUID(key.split(':')[-1]))
                found.append(aid)
        active: list[AgentId[Any]] = []
        for aid in found:
            status = await self._client.get(self._active_key(aid))
            if status == _MailboxState.ACTIVE.value:  # pragma: no branch
                active.append(aid)
        return tuple(active)

    def factory(self) -> RedisExchangeFactory:
        return RedisExchangeFactory(
            hostname=self._redis_info.hostname,
            port=self._redis_info.port,
            **self._redis_info.kwargs,
        )

    async def recv(self, timeout: float | None = None) -> Message[Any]:
        _timeout = timeout if timeout is not None else 0
        status = await self._client.get(
            self._active_key(self.mailbox_id),
        )
        if status is None:
            raise AssertionError(
                f'Status for mailbox {self.mailbox_id} did not exist in '
                'Redis server. This means that something incorrectly '
                'deleted the key.',
            )
        elif status == _MailboxState.INACTIVE.value:
            raise MailboxTerminatedError(self.mailbox_id)

        raw = await self._client.blpop(  # type: ignore[misc]
            [self._queue_key(self.mailbox_id)],
            timeout=_timeout,
        )
        if raw is None:
            raise TimeoutError(
                f'Timeout waiting for next message for {self.mailbox_id} '
                f'after {timeout} seconds.',
            )

        # Only passed one key to blpop to result is [key, item]
        assert isinstance(raw, (tuple, list))
        assert len(raw) == 2  # noqa: PLR2004
        if raw[1] == _CLOSE_SENTINEL:  # pragma: no cover
            raise MailboxTerminatedError(self.mailbox_id)
        return Message.model_deserialize(raw[1])

    async def register_agent(
        self,
        agent: type[AgentT],
        *,
        name: str | None = None,
    ) -> RedisAgentRegistration[AgentT]:
        aid: AgentId[AgentT] = AgentId.new(name=name)
        await self._client.set(
            self._active_key(aid),
            _MailboxState.ACTIVE.value,
        )
        await self._client.set(
            self._agent_key(aid),
            ','.join(agent._agent_mro()),
        )
        return RedisAgentRegistration(agent_id=aid)

    async def send(self, message: Message[Any]) -> None:
        status = await self._client.get(self._active_key(message.dest))
        if status is None:
            raise BadEntityIdError(message.dest)
        elif status == _MailboxState.INACTIVE.value:
            raise MailboxTerminatedError(message.dest)
        else:
            await self._client.rpush(  # type: ignore[misc]
                self._queue_key(message.dest),
                message.model_serialize(),
            )

    async def status(self, uid: EntityId) -> MailboxStatus:
        status = await self._client.get(self._active_key(uid))
        if status is None:
            return MailboxStatus.MISSING
        elif status == _MailboxState.INACTIVE.value:
            return MailboxStatus.TERMINATED
        else:
            return MailboxStatus.ACTIVE

    async def terminate(self, uid: EntityId) -> None:
        await self._client.set(
            self._active_key(uid),
            _MailboxState.INACTIVE.value,
        )

        pending = await self._client.lrange(self._queue_key(uid), 0, -1)  # type: ignore[misc]
        await self._client.delete(self._queue_key(uid))
        # Sending a close sentinel to the queue is a quick way to force
        # the entity waiting on messages to the mailbox to stop blocking.
        # This assumes that only one entity is reading from the mailbox.
        await self._client.rpush(self._queue_key(uid), _CLOSE_SENTINEL)  # type: ignore[misc]
        if isinstance(uid, AgentId):
            await self._client.delete(self._agent_key(uid))

        messages: list[Message[Any]] = [
            Message.model_deserialize(raw) for raw in pending
        ]
        await _respond_pending_requests_on_terminate(messages, self)


class RedisExchangeFactory(ExchangeFactory[RedisExchangeTransport]):
    """Redis exchange client factory.

    Args:
        hostname: Redis server hostname.
        port: Redis server port.
        redis_kwargs: Extra keyword arguments to pass to
            [`redis.Redis()`][redis.Redis].
    """

    def __init__(
        self,
        hostname: str,
        port: int,
        **redis_kwargs: Any,
    ) -> None:
        self.redis_info = _RedisConnectionInfo(hostname, port, redis_kwargs)

    async def _create_transport(
        self,
        mailbox_id: EntityId | None = None,
        *,
        name: str | None = None,
        registration: RedisAgentRegistration[Any] | None = None,  # type: ignore[override]
    ) -> RedisExchangeTransport:
        return await RedisExchangeTransport.new(
            mailbox_id=mailbox_id,
            name=name,
            redis_info=self.redis_info,
        )



"""Academy mypy plugin.

This plugin enables mypy to perform static type inference on
[`Handle`][academy.handle.Handle] types.

```python
from academy.agent import Agent, action
from academy.handle import Handle

class Example(Agent):
    @action
    async def get_value(self) -> int: ...

handle: Handle[Example]

reveal_type(await handle.get_value())
# note: Revealed type is "int"
```
Without the plugin, mypy will default to [`Any`][typing.Any].

Note:
    The plugin makes a best effort to warn users about incorrect use of
    agent handles. This includes raising errors when accessing attributes
    of a agent, rather than methods, via a handle and when incorrect
    parameter types are passed to an action. However, the plugin cannot
    distinguish which callable attributes of a handle are annotated as
    actions, but this will still produce an error at runtime.

Enable the plugin by adding `academy.mypy_plugin` to the list of plugins
in your
[mypy config file](https://mypy.readthedocs.io/en/latest/config_file.html){target=_blank}.

* `pyproject.toml`
  ```toml
  [tools.mypy]
  plugins = ["academy.mypy_plugin"]
  ```
* `mypy.ini` and `setup.cfg`
  ```ini
  [mypy]
  plugins = academy.mypy_plugin
  ```
"""

from __future__ import annotations

import functools
import re
import sys
from typing import Callable
from typing import TypeVar

if sys.version_info >= (3, 10):  # pragma: >=3.10 cover
    from typing import ParamSpec
else:  # pragma: <3.10 cover
    from typing_extensions import ParamSpec

from mypy.checker import TypeChecker
from mypy.errorcodes import ATTR_DEFINED
from mypy.errorcodes import UNION_ATTR
from mypy.messages import format_type
from mypy.options import Options
from mypy.plugin import AttributeContext
from mypy.plugin import MethodContext
from mypy.plugin import Plugin
from mypy.types import AnyType
from mypy.types import CallableType
from mypy.types import get_proper_type
from mypy.types import Instance
from mypy.types import NoneType
from mypy.types import Type
from mypy.types import TypeOfAny
from mypy.types import UnionType

P = ParamSpec('P')
T = TypeVar('T')


HANDLE_TYPE_PATTERN = (
    r'^academy\.handle\.([a-zA-Z_][a-zA-Z0-9_]*)?Handle'
    r'(?:\[\s*[a-zA-Z_][a-zA-Z0-9_]*\s*\])?$'
)


def is_handle_type(type_name: str) -> bool:  # noqa: D103
    # The HANDLE_TYPE_PATTERN assumes all handle types are defined in the
    # academy.handle module. If other handle types are ever introduced this
    # utility function will need to be changed to enable mypy inference.
    return bool(re.match(HANDLE_TYPE_PATTERN, type_name))


class AcademyMypyPlugin(Plugin):  # noqa: D101
    def __init__(self, options: Options) -> None:
        super().__init__(options)

    def get_attribute_hook(  # noqa: D102
        self,
        fullname: str,
    ) -> Callable[[AttributeContext], Type] | None:
        return self._get_handle_hook(fullname)

    def get_method_hook(  # noqa: D102
        self,
        fullname: str,
    ) -> Callable[[MethodContext], Type] | None:
        return self._get_handle_hook(fullname)

    def _get_handle_hook(
        self,
        fullname: str,
    ) -> Callable[[AttributeContext | MethodContext], Type] | None:
        parts = fullname.rsplit('.', 1)
        if len(parts) != 2:  # noqa: PLR2004
            return None
        type_name, attr_name = parts
        if is_handle_type(type_name):
            return functools.partial(
                handle_attr_access,
                attr_name=attr_name,
            )
        return None


def _assertion_fallback(function: Callable[P, Type]) -> Callable[P, Type]:
    # Decorator which catches AssertionErrors and returns AnyType
    # to indicate that the plugin does not know how to handle that case
    # and will default back to Any.
    # https://github.com/dry-python/returns/blob/dda187d78fe405d7d1234ffaffc99d8264f854dc/returns/contrib/mypy/_typeops/fallback.py
    @functools.wraps(function)
    def decorator(*args: P.args, **kwargs: P.kwargs) -> Type:
        try:
            return function(*args, **kwargs)
        except AssertionError:
            return AnyType(TypeOfAny.implementation_artifact)

    return decorator


def _handle_attr_access(  # noqa: C901, PLR0911, PLR0912
    instance: Type,
    attr_name: str,
    ctx: AttributeContext | MethodContext,
) -> Type:
    handle_type = get_proper_type(instance)

    if isinstance(ctx, MethodContext):
        fallback_type = ctx.default_return_type
    elif isinstance(ctx, AttributeContext):
        fallback_type = ctx.default_attr_type

    if not isinstance(handle_type, Instance):
        return fallback_type

    handle_typename = handle_type.type.fullname
    if not is_handle_type(handle_typename):
        return fallback_type

    # Attribute/method defined directly on Handle itself
    if handle_type.type.has_readable_member(attr_name):
        return fallback_type

    # After the above checks, we know instance is a Handle type and Handle
    # types are generic with one generic type.
    assert len(handle_type.args) == 1
    bound_type = get_proper_type(handle_type.args[0])
    if not isinstance(bound_type, Instance):
        # This could, for example, be a TypeVarType if we have an unbound
        # Handle[T].
        return fallback_type

    bound_attr_sym = bound_type.type.get(attr_name)
    if (
        not bound_attr_sym
        or not bound_attr_sym.node
        or not hasattr(bound_attr_sym.node, 'type')
    ):
        type_ = format_type(handle_type, ctx.api.options)
        code = ATTR_DEFINED
        if isinstance(ctx.type, UnionType):
            union = format_type(ctx.type, ctx.api.options)
            type_ = f'Item {type_} of {union}'
            code = UNION_ATTR
        ctx.api.fail(
            f'{type_} has no attribute "{attr_name}"',
            ctx.context,
            code=code,
        )
        return fallback_type

    bound_attr_type = get_proper_type(bound_attr_sym.type)
    if not isinstance(bound_attr_type, CallableType):
        if bound_attr_type is not None:
            handle_type_ = format_type(handle_type, ctx.api.options)
            bound_type_ = format_type(bound_type, ctx.api.options)
            code = ATTR_DEFINED
            if isinstance(ctx.type, UnionType):
                union = format_type(ctx.type, ctx.api.options)
                handle_type_ = f'Item {handle_type_} of {union}'
                code = UNION_ATTR
            ctx.api.fail(
                f'{handle_type_} has no method "{attr_name}"; '
                f'only action-decorated methods of {bound_type_} '
                'can accessed via a handle',
                ctx.context,
                code=code,
            )
        return fallback_type

    ret_type: Type
    ret_ret_type = get_proper_type(bound_attr_type.ret_type)
    if (
        isinstance(ret_ret_type, Instance)
        and ret_ret_type.type.fullname == 'typing.Coroutine'
        and len(ret_ret_type.args) == 3  # noqa: PLR2004 (Coroutine[A, B, C])
    ):
        inner_type = ret_ret_type.args[2]
        assert isinstance(ctx.api, TypeChecker)
        coroutine_type = ctx.api.named_type('typing.Coroutine')
        ret_type = coroutine_type.copy_modified(
            # Coroutine[None, None, T]
            args=[NoneType(), NoneType(), inner_type],
        )
    else:
        return fallback_type

    if isinstance(ctx, AttributeContext):
        # Need to drop the "self" argument so that mypy does not think
        # it was forgotten.
        return bound_attr_type.copy_modified(
            arg_types=bound_attr_type.arg_types[1:],
            arg_kinds=bound_attr_type.arg_kinds[1:],
            arg_names=bound_attr_type.arg_names[1:],
            ret_type=ret_type,
        )
    elif isinstance(ctx, MethodContext):
        return ret_type


@_assertion_fallback
def handle_attr_access(  # noqa: D103
    ctx: AttributeContext | MethodContext,
    attr_name: str,
) -> Type:
    if isinstance(ctx.type, UnionType):
        resolved = tuple(
            _handle_attr_access(instance, attr_name, ctx)
            for instance in ctx.type.items
        )
        return UnionType(resolved)
    else:
        return _handle_attr_access(ctx.type, attr_name, ctx)


def plugin(version: str) -> type[AcademyMypyPlugin]:  # noqa: D103
    return AcademyMypyPlugin



from __future__ import annotations

from academy.exchange.client import AgentExchangeClient
from academy.exchange.client import ExchangeClient
from academy.exchange.client import UserExchangeClient
from academy.exchange.cloud import HttpExchangeFactory
from academy.exchange.cloud import HttpExchangeTransport
from academy.exchange.factory import ExchangeFactory
from academy.exchange.hybrid import HybridExchangeFactory
from academy.exchange.hybrid import HybridExchangeTransport
from academy.exchange.local import LocalExchangeFactory
from academy.exchange.local import LocalExchangeTransport
from academy.exchange.proxystore import ProxyStoreExchangeFactory
from academy.exchange.proxystore import ProxyStoreExchangeTransport
from academy.exchange.redis import RedisExchangeFactory
from academy.exchange.redis import RedisExchangeTransport
from academy.exchange.transport import ExchangeTransport
from academy.exchange.transport import MailboxStatus

__all__ = [
    'AgentExchangeClient',
    'ExchangeClient',
    'ExchangeFactory',
    'ExchangeTransport',
    'HttpExchangeFactory',
    'HttpExchangeTransport',
    'HybridExchangeFactory',
    'HybridExchangeTransport',
    'LocalExchangeFactory',
    'LocalExchangeTransport',
    'MailboxStatus',
    'ProxyStoreExchangeFactory',
    'ProxyStoreExchangeTransport',
    'RedisExchangeFactory',
    'RedisExchangeTransport',
    'UserExchangeClient',
]



from __future__ import annotations

import pickle

import pytest

from academy.exchange.local import LocalExchangeFactory


def test_factory_serialize_error(
    local_exchange_factory: LocalExchangeFactory,
) -> None:
    with pytest.raises(pickle.PicklingError):
        pickle.dumps(local_exchange_factory)



from __future__ import annotations

import pathlib

from academy.exchange.cloud.config import ExchangeAuthConfig
from academy.exchange.cloud.config import ExchangeServingConfig


def test_auth_config_default() -> None:
    config = ExchangeAuthConfig()
    assert config.method is None


def test_read_from_config_file_empty(tmp_path: pathlib.Path) -> None:
    data = '[serving]'

    filepath = tmp_path / 'relay.toml'
    with open(filepath, 'w') as f:
        f.write(data)

    config = ExchangeServingConfig.from_toml(filepath)
    assert config == ExchangeServingConfig()


def test_read_from_config_file(tmp_path: pathlib.Path) -> None:
    data = """\
host = "localhost"
port = 1234
certfile = "/path/to/cert.pem"
keyfile = "/path/to/privkey.pem"

[auth]
method = "globus"

[auth.kwargs]
client_id = "ABC"
"""

    filepath = tmp_path / 'relay.toml'
    with open(filepath, 'w') as f:
        f.write(data)

    config = ExchangeServingConfig.from_toml(filepath)

    assert config.host == 'localhost'
    assert config.port == 1234  # noqa: PLR2004
    assert config.certfile == '/path/to/cert.pem'
    assert config.keyfile == '/path/to/privkey.pem'

    assert config.auth.method == 'globus'
    assert config.auth.kwargs['client_id'] == 'ABC'
    assert 'client_secret' not in config.auth.kwargs



from __future__ import annotations

import sys
from collections.abc import Iterable
from typing import Any

from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.identifier import UserId


class ActionCancelledError(Exception):
    """Action was cancelled by the agent.

    This often happens when an agent is shutdown mid-action execution and
    configured to cancel running actions.
    """

    def __init__(self, name: str) -> None:
        super().__init__(f'Action "{name}" was cancelled by the agent.')


class AgentNotInitializedError(Exception):
    """Agent runtime context has not been initialized.

    This error is typically raised when accessing the runtime context for
    an agent before the agent has been executed.
    """

    def __init__(self) -> None:
        super().__init__(
            'Agent runtime context has not been initialized. '
            'Has the agent been started?',
        )


class ExchangeError(Exception):
    """Base type for exchange related errors."""

    pass


class BadEntityIdError(ExchangeError):
    """Entity associated with the identifier is unknown."""

    def __init__(self, uid: EntityId) -> None:
        super().__init__(f'Unknown identifier {uid}.')


class ForbiddenError(ExchangeError):
    """Exchange client does not have permission to access resources."""

    pass


class MailboxTerminatedError(ExchangeError):
    """Entity mailbox is terminated and cannot send or receive messages.

    Constructing this error type implicitly returns one of the derived types,
    [`AgentTerminatedError`][academy.exception.AgentTerminatedError] or
    [`UserTerminatedError`][academy.exception.UserTerminatedError], based
    on the entity type.
    """

    def __new__(cls, uid: EntityId) -> MailboxTerminatedError:  # noqa: D102
        if isinstance(uid, AgentId):
            return super().__new__(AgentTerminatedError)
        elif isinstance(uid, UserId):
            return super().__new__(UserTerminatedError)
        else:
            raise AssertionError('Unreachable.')

    def __init__(self, uid: EntityId) -> None:
        super().__init__(f'Mailbox for {uid} has been terminated.')
        self.uid = uid

    def __reduce__(self) -> Any:
        # BaseException implements __reduce__ as
        #     return type(self), self.args
        # where args will contain the message passed to super().__init__
        # rather than uid so it must be customized.
        return type(self), (self.uid,)


class AgentTerminatedError(MailboxTerminatedError):
    """Agent mailbox is terminated and cannot send or receive messages."""

    def __init__(self, uid: AgentId[Any]) -> None:
        super().__init__(uid)


class UserTerminatedError(MailboxTerminatedError):
    """User mailbox is terminated and cannot send or receive messages."""

    def __init__(self, uid: UserId) -> None:
        super().__init__(uid)


class UnauthorizedError(ExchangeError):
    """Exchange client has not provided valid authentication credentials."""

    pass


class ExchangeClientNotFoundError(Exception):
    """Handle to agent can not find an exchange client to use.

    A [`Handle`][academy.handle.Handle] is
    initialized with a target agent ID is not used in a context where an
    exchange client could be inferred. Typically this can be resolved by
    using a [`ExchangeClient`][academy.exchange.ExchangeClient] or
    [`Manager`][academy.manager.Manager] as a context manager. If this error
    happens within an agent, it likely means the agent was not started.
    """

    def __init__(self, aid: AgentId[Any]) -> None:
        super().__init__(
            f'Handle to {aid} can not find an exchange client to use. See the '
            'exception docstring for troubleshooting.',
        )


def raise_exceptions(
    exceptions: Iterable[BaseException],
    *,
    message: str | None = None,
) -> None:
    """Raise exceptions as a group.

    Raises a set of exceptions as an [`ExceptionGroup`][ExceptionGroup]
    in Python 3.11 and later. If only one exception is provided, it is raised
    directly. In Python 3.10 and older, only one exception is raised.

    This is a no-op if the size of `exceptions` is zero.

    Args:
        exceptions: An iterable of exceptions to raise.
        message: Custom error message for the exception group.
    """
    excs = tuple(exceptions)
    if len(excs) == 0:
        return

    if sys.version_info >= (3, 11) and len(excs) > 1:  # pragma: >=3.11 cover
        message = (
            message if message is not None else 'Caught multiple exceptions!'
        )
        # Note that BaseExceptionGroup will return ExceptionGroup if all
        # of the errors are Exception, rather than BaseException, so that this
        # can be caught by "except Exception".
        raise BaseExceptionGroup(message, excs)  # noqa: F821
    else:  # pragma: <3.11 cover
        raise excs[0]



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor

import pytest

from academy.agent import action
from academy.agent import Agent
from academy.exchange import LocalExchangeFactory
from academy.manager import Manager


class Counter(Agent):
    def __init__(self) -> None:
        super().__init__()
        self.count = 0

    @action
    async def increment(self, value: int = 1) -> None:
        self.count += value

    @action
    async def get_count(self) -> int:
        return self.count


@pytest.mark.asyncio
async def test_user_agent_handles() -> None:
    async with await Manager.from_exchange_factory(
        factory=LocalExchangeFactory(),
        executors=ThreadPoolExecutor(),
    ) as manager:
        agent_handle = await manager.launch(Counter)

        assert await agent_handle.ping() > 0

        count = await agent_handle.get_count()
        assert count == 0

        await agent_handle.increment()

        count = await agent_handle.get_count()
        assert count == 1

        await agent_handle.shutdown()
        await manager.wait({agent_handle}, timeout=1)



from __future__ import annotations

import asyncio
import sys
from typing import Any
from unittest import mock

import pytest

from academy.agent import action
from academy.agent import Agent
from academy.agent import loop
from academy.context import ActionContext
from academy.exception import ActionCancelledError
from academy.exchange import LocalExchangeTransport
from academy.exchange import UserExchangeClient
from academy.exchange.transport import MailboxStatus
from academy.handle import Handle
from academy.handle import ProxyHandle
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.message import ActionRequest
from academy.message import ActionResponse
from academy.message import ErrorResponse
from academy.message import Message
from academy.message import PingRequest
from academy.message import ShutdownRequest
from academy.message import SuccessResponse
from academy.runtime import Runtime
from academy.runtime import RuntimeConfig
from testing.agents import CounterAgent
from testing.agents import EmptyAgent
from testing.agents import ErrorAgent
from testing.constant import TEST_SLEEP_INTERVAL
from testing.constant import TEST_WAIT_TIMEOUT


class SignalingAgent(Agent):
    def __init__(self) -> None:
        super().__init__()
        self.startup_event = asyncio.Event()
        self.shutdown_event = asyncio.Event()

    async def agent_on_startup(self) -> None:
        self.startup_event.set()

    async def agent_on_shutdown(self) -> None:
        self.shutdown_event.set()

    @loop
    async def shutdown_immediately(self, shutdown: asyncio.Event) -> None:
        shutdown.set()


@pytest.mark.asyncio
async def test_runtime_context_manager(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(SignalingAgent)
    async with Runtime(
        EmptyAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    ) as runtime:
        assert isinstance(repr(runtime), str)
        assert isinstance(str(runtime), str)


@pytest.mark.asyncio
async def test_runtime_run_until_complete(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(SignalingAgent)
    runtime = Runtime(
        SignalingAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    )
    await runtime.run_until_complete()

    with pytest.raises(RuntimeError, match='Agent has already been shutdown.'):
        await runtime.run_until_complete()

    assert runtime.agent.startup_event.is_set()
    assert runtime.agent.shutdown_event.is_set()


@pytest.mark.asyncio
async def test_runtime_run_until_complete_as_task(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(SignalingAgent)
    runtime = Runtime(
        SignalingAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    )

    task = asyncio.create_task(
        runtime.run_until_complete(),
        name='test-runtime-run-until-complete-at-task',
    )
    await task

    assert runtime.agent.startup_event.is_set()
    assert runtime.agent.shutdown_event.is_set()


@pytest.mark.asyncio
async def test_runtime_shutdown_without_terminate(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(SignalingAgent)
    runtime = Runtime(
        SignalingAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
        config=RuntimeConfig(terminate_on_success=False),
    )
    await runtime.run_until_complete()
    assert runtime._shutdown_options.expected_shutdown
    assert (
        await exchange_client.status(runtime.agent_id) == MailboxStatus.ACTIVE
    )


@pytest.mark.asyncio
async def test_runtime_shutdown_terminate_override(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(EmptyAgent)

    async with Runtime(
        EmptyAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
        config=RuntimeConfig(
            terminate_on_success=False,
            terminate_on_error=False,
        ),
    ) as runtime:
        runtime.signal_shutdown(expected=True, terminate=True)
        await runtime.wait_shutdown(timeout=TEST_WAIT_TIMEOUT)

    assert (
        await exchange_client.status(runtime.agent_id)
        == MailboxStatus.TERMINATED
    )


@pytest.mark.asyncio
async def test_runtime_startup_failure(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(SignalingAgent)
    runtime = Runtime(
        SignalingAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    )

    with mock.patch.object(runtime, '_start', side_effect=Exception('Oops!')):
        with pytest.raises(Exception, match='Oops!'):
            await runtime.run_until_complete()

    assert not runtime.agent.startup_event.is_set()
    assert not runtime.agent.shutdown_event.is_set()


@pytest.mark.asyncio
async def test_runtime_wait_shutdown_timeout(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(SignalingAgent)
    async with Runtime(
        EmptyAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    ) as runtime:
        with pytest.raises(TimeoutError):
            await runtime.wait_shutdown(timeout=TEST_SLEEP_INTERVAL)


class LoopFailureAgent(Agent):
    @loop
    async def bad1(self, shutdown: asyncio.Event) -> None:
        raise RuntimeError('Loop failure 1.')

    @loop
    async def bad2(self, shutdown: asyncio.Event) -> None:
        raise RuntimeError('Loop failure 2.')


@pytest.mark.parametrize('raise_errors', (True, False))
@pytest.mark.asyncio
async def test_runtime_loop_error_causes_shutdown(
    raise_errors: bool,
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(LoopFailureAgent)
    runtime = Runtime(
        LoopFailureAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
        config=RuntimeConfig(raise_loop_errors_on_shutdown=raise_errors),
    )

    if not raise_errors:
        await asyncio.wait_for(
            runtime.run_until_complete(),
            timeout=TEST_WAIT_TIMEOUT,
        )
    elif sys.version_info >= (3, 11):  # pragma: >=3.11 cover
        # In Python 3.11 and later, all exceptions are raised in a group.
        with pytest.raises(ExceptionGroup) as exc_info:  # noqa: F821
            await asyncio.wait_for(
                runtime.run_until_complete(),
                timeout=TEST_WAIT_TIMEOUT,
            )
        assert len(exc_info.value.exceptions) == 2  # noqa: PLR2004
    else:  # pragma: <3.11 cover
        # In Python 3.10 and older, only the first error will be raised.
        with pytest.raises(RuntimeError, match='Loop failure'):
            await asyncio.wait_for(
                runtime.run_until_complete(),
                timeout=TEST_WAIT_TIMEOUT,
            )


@pytest.mark.asyncio
async def test_runtime_loop_error_without_shutdown(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(LoopFailureAgent)
    runtime = Runtime(
        LoopFailureAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
        config=RuntimeConfig(shutdown_on_loop_error=False),
    )

    task = asyncio.create_task(
        runtime.run_until_complete(),
        name='test-runtime-loop-error-without-shutdown',
    )
    await runtime._started_event.wait()

    # Should timeout because agent did not shutdown after loop errors
    done, pending = await asyncio.wait({task}, timeout=TEST_SLEEP_INTERVAL)
    assert len(done) == 0
    runtime.signal_shutdown()

    # Loop errors raised on shutdown
    if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
        # In Python 3.11 and later, all exceptions are raised in a group.
        with pytest.raises(ExceptionGroup) as exc_info:  # noqa: F821
            await task
        assert len(exc_info.value.exceptions) == 2  # noqa: PLR2004
    else:  # pragma: <3.11 cover
        # In Python 3.10 and older, only the first error will be raised.
        with pytest.raises(RuntimeError, match='Loop failure'):
            await task


@pytest.mark.asyncio
async def test_runtime_shutdown_message(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(EmptyAgent)

    async with Runtime(
        EmptyAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    ) as runtime:
        shutdown = Message.create(
            src=exchange_client.client_id,
            dest=runtime.agent_id,
            body=ShutdownRequest(),
        )
        await exchange_client.send(shutdown)
        await runtime.wait_shutdown(timeout=TEST_WAIT_TIMEOUT)


class InfiniteAgent(Agent):
    @loop
    async def wait(self, shutdown: asyncio.Event) -> None:
        while True:
            await asyncio.sleep(5)


@pytest.mark.asyncio
async def test_runtime_cancel_loop(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(InfiniteAgent)

    runtime = Runtime(
        InfiniteAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    )

    task = asyncio.create_task(
        runtime.run_until_complete(),
        name='test-runtime-cancel-loop',
    )
    await runtime._started_event.wait()

    # Should cancel loop without raising error
    runtime.signal_shutdown()
    await task


@pytest.mark.asyncio
async def test_runtime_ping_message(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(EmptyAgent)
    # Cancel listener so test can intercept agent responses
    await exchange_client._stop_listener_task()

    async with Runtime(
        EmptyAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    ) as runtime:
        ping = Message.create(
            src=exchange_client.client_id,
            dest=runtime.agent_id,
            body=PingRequest(),
        )
        await exchange_client.send(ping)
        message = await exchange_client._transport.recv()
        assert isinstance(message.get_body(), SuccessResponse)


@pytest.mark.asyncio
async def test_runtime_action_message(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(CounterAgent)
    # Cancel listener so test can intercept agent responses
    await exchange_client._stop_listener_task()

    async with Runtime(
        CounterAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    ) as runtime:
        value = 42
        request = Message.create(
            src=exchange_client.client_id,
            dest=runtime.agent_id,
            body=ActionRequest(action='add', pargs=(value,)),
        )
        await exchange_client.send(request)
        message = await exchange_client._transport.recv()
        body = message.get_body()
        assert isinstance(body, ActionResponse)
        assert body.get_result() is None

        request = Message.create(
            src=exchange_client.client_id,
            dest=runtime.agent_id,
            body=ActionRequest(action='count'),
        )
        await exchange_client.send(request)
        message = await exchange_client._transport.recv()
        body = message.get_body()
        assert isinstance(body, ActionResponse)
        assert body.get_result() == value


@pytest.mark.parametrize('cancel', (True, False))
@pytest.mark.asyncio
async def test_runtime_cancel_action_requests_on_shutdown(
    cancel: bool,
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    class NoReturnAgent(Agent):
        @action
        async def sleep(self) -> None:
            await asyncio.sleep(1000 if cancel else TEST_SLEEP_INTERVAL)

    registration = await exchange_client.register_agent(ErrorAgent)
    # Cancel listener so test can intercept agent responses
    await exchange_client._stop_listener_task()

    runtime = Runtime(
        NoReturnAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
        config=RuntimeConfig(cancel_actions_on_shutdown=cancel),
    )
    task = asyncio.create_task(
        runtime.run_until_complete(),
        name='test-runtime-cancel-action-requests-on-shutdown',
    )
    await runtime._started_event.wait()

    request = Message.create(
        src=exchange_client.client_id,
        dest=runtime.agent_id,
        body=ActionRequest(action='sleep'),
    )
    await exchange_client.send(request)

    shutdown = Message.create(
        src=exchange_client.client_id,
        dest=runtime.agent_id,
        body=ShutdownRequest(),
    )
    await exchange_client.send(shutdown)

    message = await exchange_client._transport.recv()
    body = message.get_body()
    if cancel:
        assert isinstance(body, ErrorResponse)
        assert isinstance(body.get_exception(), ActionCancelledError)
    else:
        assert isinstance(body, ActionResponse)
        assert body.get_result() is None

    await asyncio.wait_for(task, timeout=TEST_WAIT_TIMEOUT)


@pytest.mark.asyncio
async def test_runtime_action_message_error(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(ErrorAgent)
    # Cancel listener so test can intercept agent responses
    await exchange_client._stop_listener_task()

    async with Runtime(
        ErrorAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    ) as runtime:
        request = Message.create(
            src=exchange_client.client_id,
            dest=runtime.agent_id,
            body=ActionRequest(action='fails'),
        )
        await exchange_client.send(request)
        message = await exchange_client._transport.recv()
        body = message.get_body()
        assert isinstance(body, ErrorResponse)
        exception = body.get_exception()
        assert isinstance(exception, RuntimeError)
        assert 'This action always fails.' in str(exception)


@pytest.mark.asyncio
async def test_runtime_action_message_unknown(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(EmptyAgent)
    # Cancel listener so test can intercept agent responses
    await exchange_client._stop_listener_task()

    async with Runtime(
        EmptyAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    ) as runtime:
        request = Message.create(
            src=exchange_client.client_id,
            dest=runtime.agent_id,
            body=ActionRequest(action='null'),
        )
        await exchange_client.send(request)
        message = await exchange_client._transport.recv()
        body = message.get_body()
        assert isinstance(body, ErrorResponse)
        exception = body.get_exception()
        assert isinstance(exception, AttributeError)
        assert 'null' in str(exception)


@pytest.mark.asyncio
async def test_runtime_delay_actions_and_loops_to_after_startup(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    class ExampleAgent(Agent):
        def __init__(self) -> None:
            self.startup_called = False

        async def agent_on_startup(self) -> None:
            # Simulate some work that yields execution to other scheduled
            # tasks so that the scheduled action tasks gets run concurrently
            # with the startup callback. The action task should wait
            # on the startup sequence to finish and immediately yield back
            # control so the callback can finish
            for _ in range(10):
                await asyncio.sleep(0)
            self.startup_called = True

        @action
        async def check_action(self) -> None:
            assert self.startup_called

        @loop
        async def check_loop(self, shutdown: asyncio.Event) -> None:
            assert self.startup_called

    registration = await exchange_client.register_agent(ExampleAgent)
    # Cancel listener so test can intercept agent responses
    await exchange_client._stop_listener_task()

    # Send action request before starting agent so its immediately
    # available when message listener task starts
    request = Message.create(
        src=exchange_client.client_id,
        dest=registration.agent_id,
        body=ActionRequest(action='check_action'),
    )
    await exchange_client.send(request)

    async with Runtime(
        ExampleAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    ):
        message = await exchange_client._transport.recv()
        body = message.get_body()
        assert isinstance(body, ActionResponse)
        assert body.get_result() is None


@pytest.mark.asyncio
async def test_agent_exchange_context(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    class _TestAgent(Agent):
        def __init__(
            self,
            handle: Handle[EmptyAgent],
            proxy: ProxyHandle[EmptyAgent],
        ) -> None:
            super().__init__()
            self.direct = handle
            self.proxy = proxy
            self.sequence = [handle]
            self.mapping = {'x': handle}

    factory = exchange_client.factory()
    registration = await exchange_client.register_agent(_TestAgent)
    proxy_handle = ProxyHandle(EmptyAgent())
    unbound_handle = Handle(
        (await exchange_client.register_agent(EmptyAgent)).agent_id,
    )

    async def _request_handler(_: Any) -> None:  # pragma: no cover
        pass

    async with await factory.create_agent_client(
        registration,
        _request_handler,
    ) as agent_client:
        agent = _TestAgent(unbound_handle, proxy_handle)
        assert agent.proxy is proxy_handle
        assert isinstance(agent.direct, Handle)
        assert agent.direct.exchange is agent_client
        for handle in agent.sequence:
            assert isinstance(handle, Handle)
            assert handle.exchange is agent_client
        for handle in agent.mapping.values():
            assert isinstance(handle, Handle)
            assert handle.exchange is agent_client


class ShutdownAgent(Agent):
    @action
    async def end(self) -> None:
        self.agent_shutdown()


@pytest.mark.asyncio
async def test_runtime_agent_self_termination(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(ShutdownAgent)

    async with Runtime(
        ShutdownAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    ) as runtime:
        await runtime.action('end', AgentId.new(), args=(), kwargs={})
        await runtime.wait_shutdown(timeout=TEST_WAIT_TIMEOUT)


class ContextAgent(Agent):
    @action(context=True)
    async def call(
        self,
        source_id: EntityId,
        *,
        context: ActionContext,
    ) -> None:
        assert source_id == context.source_id


@pytest.mark.asyncio
async def test_runtime_agent_action_context(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    registration = await exchange_client.register_agent(ShutdownAgent)

    async with Runtime(
        ContextAgent(),
        exchange_factory=exchange_client.factory(),
        registration=registration,
    ) as runtime:
        await runtime.action(
            'call',
            exchange_client.client_id,
            args=(exchange_client.client_id,),
            kwargs={},
        )



from __future__ import annotations

from typing import Any

import pytest

from academy.context import ActionContext
from academy.exchange import LocalExchangeTransport
from academy.exchange import UserExchangeClient
from academy.identifier import AgentId
from academy.identifier import UserId
from testing.agents import EmptyAgent


@pytest.mark.asyncio
async def test_action_context_agent_source(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    factory = exchange_client.factory()
    source_id: AgentId[EmptyAgent] = AgentId.new()
    registration = await exchange_client.register_agent(EmptyAgent)

    async def _request_handler(_: Any) -> None:  # pragma: no cover
        pass

    async with await factory.create_agent_client(
        registration,
        _request_handler,
    ) as agent_client:
        context = ActionContext(source_id, agent_client)
        assert context.is_agent_source()
        assert not context.is_user_source()
        handle = context.source_handle
        assert handle.agent_id == source_id
        # Check that handle.source_handle is cached
        assert context.source_handle is handle


@pytest.mark.asyncio
async def test_action_context_user_source(
    exchange_client: UserExchangeClient[LocalExchangeTransport],
) -> None:
    factory = exchange_client.factory()
    source_id = UserId.new()
    registration = await exchange_client.register_agent(EmptyAgent)

    async def _request_handler(_: Any) -> None:  # pragma: no cover
        pass

    async with await factory.create_agent_client(
        registration,
        _request_handler,
    ) as agent_client:
        context = ActionContext(source_id, agent_client)
        assert not context.is_agent_source()
        assert context.is_user_source()

        with pytest.raises(TypeError):
            _ = context.source_handle



from __future__ import annotations

import logging
import pathlib

import pytest

from academy.logging import init_logging

# Note: these tests are just for coverage to make sure the code is functional.
# It does not test the agent of init_logging because pytest captures
# logging already.


@pytest.mark.parametrize(('color', 'extra'), ((True, True), (False, False)))
def test_logging_no_file(color: bool, extra: bool) -> None:
    init_logging(color=color, extra=extra)

    logger = logging.getLogger()
    logger.info('Test logging')


@pytest.mark.parametrize(('color', 'extra'), ((True, True), (False, False)))
def test_logging_with_file(
    color: bool,
    extra: bool,
    tmp_path: pathlib.Path,
) -> None:
    filepath = tmp_path / 'log.txt'
    init_logging(logfile=filepath, color=color, extra=extra)

    logger = logging.getLogger()
    logger.info('Test logging')



from __future__ import annotations

import asyncio
import contextlib
import fcntl
import logging
import platform
import socket
import struct
import sys
import time
from collections.abc import AsyncGenerator
from collections.abc import Awaitable
from contextlib import asynccontextmanager
from types import TracebackType
from typing import Callable

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

logger = logging.getLogger(__name__)

_BAD_FILE_DESCRIPTOR_ERRNO = 9
MESSAGE_CHUNK_SIZE = 10240
MESSAGE_HEADER_FORMAT = '!I'  # Network byte order, unsigned 4-byte int
MESSAGE_HEADER_SIZE = struct.calcsize(MESSAGE_HEADER_FORMAT)


class SocketClosedError(Exception):
    """Socket is already closed."""

    pass


class SocketOpenError(Exception):
    """Failed to open socket."""

    pass


class SimpleSocket:
    """Simple socket wrapper.

    Configures a client connection using a non-blocking TCP socket over IPv4.
    The send and recv methods handle byte encoding, message delimiters, and
    partial message buffering.

    Note:
        This class can be used as an async context manager.

    Args:
        reader: Socket reader interface.
        writer: Socket writer interface.
        timeout: Optional timeout for socket operations.
    """

    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        *,
        timeout: float | None = None,
    ) -> None:
        self.reader = reader
        self.writer = writer
        self.timeout = timeout
        self.closed = False

    @classmethod
    async def connect(
        cls,
        host: str,
        port: int,
        *,
        timeout: float | None = None,
    ) -> Self:
        """Establish a new TCP connection.

        Args:
            host: Host address to connect to.
            port: Port to connect to.
            timeout: Connection establish timeout.

        Raises:
            SocketOpenError: If creating the socket fails. The `__cause__` of
                the exception will be set to the underlying `OSError`.
        """
        try:
            coro = asyncio.open_connection(host, port)
            reader, writer = await asyncio.wait_for(coro, timeout)
        except (OSError, asyncio.TimeoutError) as e:
            raise SocketOpenError() from e

        return cls(reader, writer, timeout=timeout)

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        await self.close()

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.reader!r}, {self.writer!r})'

    def __str__(self) -> str:
        return f'{type(self).__name__}<{self.reader}, {self.writer}>'

    async def close(self, shutdown: bool = True) -> None:
        """Close the socket.

        Args:
            shutdown: Write EOF to the socket.
        """
        if self.closed:
            return
        self.closed = True
        if shutdown and not self.writer.is_closing():
            self.writer.write_eof()
        self.writer.close()
        await self.writer.wait_closed()

    async def send(self, message: bytes) -> None:
        """Send bytes to the socket.

        Note:
            This is a noop if the message is empty.

        Args:
            message: Message to send.

        Raises:
            SocketClosedError: If the socket was closed.
            OSError: If an error occurred.
        """
        if self.closed or self.writer.is_closing():
            raise SocketClosedError()

        if not message:
            return

        header = _make_header(message)
        self.writer.write(header)
        await self.writer.drain()

        sent_size = 0
        message_size = len(message)
        while sent_size < message_size:
            nbytes = min(message_size - sent_size, MESSAGE_CHUNK_SIZE)
            chunk = message[sent_size : sent_size + nbytes]
            self.writer.write(chunk)
            await self.writer.drain()
            sent_size += len(chunk)

    async def send_string(self, message: str) -> None:
        """Send a string to the socket.

        Strings are encoded with UTF-8.

        Args:
            message: Message to send.

        Raises:
            SocketClosedError: if the socket was closed.
            OSError: if an error occurred.
        """
        await self.send(message.encode('utf-8'))

    async def recv(self) -> bytes:
        """Receive the next message from the socket.

        Returns:
            Bytes containing the message.

        Raises:
            SocketClosedError: if the socket was closed.
            OSError: if an error occurred.
        """
        if self.closed:
            raise SocketClosedError()

        header = await self.reader.readexactly(MESSAGE_HEADER_SIZE)
        message_size = _get_size_from_header(header)

        buffer = bytearray(message_size)
        received = 0
        while received < message_size:
            nbytes = min(message_size - received, MESSAGE_CHUNK_SIZE)
            chunk = await self.reader.readexactly(nbytes)
            buffer[received : received + len(chunk)] = chunk
            received += len(chunk)

        return buffer

    async def recv_string(self) -> str:
        """Receive the next message from the socket.

        Returns:
            Message decoded as a UTF-8 string.

        Raises:
            SocketClosedError: if the socket was closed.
            OSError: if an error occurred.
        """
        return (await self.recv()).decode('utf-8')


class SocketPool:
    """Simple socket pool."""

    def __init__(self) -> None:
        self._sockets: dict[str, SimpleSocket] = {}
        self._lock = asyncio.Lock()

    async def close(self) -> None:
        """Close all the sockets in the pool."""
        for address in tuple(self._sockets.keys()):
            await self.close_socket(address)

    async def close_socket(self, address: str) -> None:
        """Close the socket for the given address."""
        async with self._lock:
            conn = self._sockets.pop(address, None)
        if conn is not None:  # pragma: no branch
            await conn.close(shutdown=True)

    async def get_socket(self, address: str) -> SimpleSocket:
        """Get or create a socket for a given address."""
        async with self._lock:
            try:
                return self._sockets[address]
            except KeyError:
                parts = address.split(':')
                host, port = parts[0], int(parts[1])
                conn = await SimpleSocket.connect(host, port)
                self._sockets[address] = conn
                return conn

    async def send(self, address: str, message: bytes) -> None:
        """Send a message to a given address."""
        conn = await self.get_socket(address)
        try:
            await conn.send(message)
        except (SocketClosedError, OSError):
            await self.close_socket(address)
            raise


class SimpleSocketServer:
    """Simple asyncio TCP socket server.

    Args:
        handler: Callback that handles a message and returns the response
            string. The handler is called synchronously within the client
            handler so it should not perform any heavy/blocking operations.
        host: Host to bind to.
        port: Port to bind to. If `None`, a random port is bound to.
    """

    def __init__(
        self,
        handler: Callable[[bytes], Awaitable[bytes | None]],
        *,
        host: str = '0.0.0.0',
        port: int | None = None,
    ) -> None:
        self.host = host
        self.port = port if port is not None else open_port()
        self.handler = handler
        self._client_tasks: set[asyncio.Task[None]] = set()

    async def _read_message(
        self,
        reader: asyncio.StreamReader,
    ) -> bytes | bytearray:
        header = await reader.read(MESSAGE_HEADER_SIZE)
        if len(header) == 0:  # pragma: no cover
            return b''
        message_size = _get_size_from_header(header)

        buffer = bytearray(message_size)
        received = 0
        while received < message_size:
            nbytes = min(message_size - received, MESSAGE_CHUNK_SIZE)
            chunk = await reader.read(nbytes)
            # buffer.extend(chunk)
            buffer[received : received + len(chunk)] = chunk
            received += len(chunk)

        return buffer

    async def _write_message(
        self,
        writer: asyncio.StreamWriter,
        message: bytes | bytearray,
    ) -> None:
        message_size = len(message)
        header = _make_header(message)
        writer.write(header)

        sent_size = 0
        while sent_size < message_size:
            nbytes = min(message_size - sent_size, MESSAGE_CHUNK_SIZE)
            chunk = message[sent_size : sent_size + nbytes]
            writer.write(chunk)
            await writer.drain()
            sent_size += len(chunk)

        await writer.drain()

    async def _handle_client(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        try:
            while not reader.at_eof():  # pragma: no branch
                try:
                    message = await self._read_message(reader)
                except asyncio.IncompleteReadError:  # pragma: no cover
                    reader.feed_eof()
                    continue
                else:
                    if len(message) == 0:  # pragma: no cover
                        break
                    response = await self.handler(message)
                    if response is not None:  # pragma: no break
                        await self._write_message(writer, response)
        except Exception:  # pragma: no cover
            logger.exception('Error in client handler task.')
            raise
        finally:
            writer.close()
            await writer.wait_closed()

    async def _register_client_task(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        task = asyncio.create_task(self._handle_client(reader, writer))
        self._client_tasks.add(task)
        task.add_done_callback(self._client_tasks.discard)

    @asynccontextmanager
    async def serve(self) -> AsyncGenerator[Self]:
        """Serve in a context manager."""
        server = await asyncio.start_server(
            self._register_client_task,
            host=self.host,
            port=self.port,
        )
        logger.debug('TCP server listening at %s:%s', self.host, self.port)

        async with server:
            await server.start_serving()

            yield self

            for task in tuple(self._client_tasks):
                task.cancel('Server has been closed.')
                with contextlib.suppress(asyncio.CancelledError):
                    await task

        if sys.version_info >= (3, 13):  # pragma: >=3.13 cover
            server.close_clients()
        logger.debug('TCP server finished at %s:%s', self.host, self.port)


def _get_size_from_header(header: bytes) -> int:
    return struct.unpack(MESSAGE_HEADER_FORMAT, header)[0]


def _make_header(message: bytes | bytearray) -> bytes:
    return struct.pack(MESSAGE_HEADER_FORMAT, len(message))


def address_by_hostname() -> str:
    """Get the IP address from the hostname of the local host."""
    return socket.gethostbyname(platform.node())


def address_by_interface(ifname: str) -> str:  # pragma: darwin no cover
    """Get the IP address of the given interface.

    Source: https://stackoverflow.com/questions/24196932/how-can-i-get-the-ip-address-of-eth0-in-python#24196955

    Args:
        ifname: Name of the interface whose address is to be returned.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(
        fcntl.ioctl(
            s.fileno(),
            0x8915,  # SIOCGIFADDR
            struct.pack('256s', bytes(ifname[:15], 'utf-8')),
        )[20:24],
    )


_used_ports: set[int] = set()


def open_port() -> int:
    """Return open port.

    Source: https://stackoverflow.com/questions/2838244
    """
    while True:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(('', 0))
        s.listen(1)
        port = s.getsockname()[1]
        s.close()
        if port not in _used_ports:  # pragma: no branch
            _used_ports.add(port)
            return port


def wait_connection(
    host: str,
    port: int,
    *,
    sleep: float = 0.01,
    timeout: float | None = None,
) -> None:
    """Wait for a socket connection to be established.

    Repeatedly tries to open and close a socket connection to `host:port`.
    If successful, the function returns. If unsuccessful before the timeout,
    a `TimeoutError` is raised. The function will sleep for `sleep` seconds
    in between successive connection attempts.

    Args:
        host: Host address to connect to.
        port: Host port to connect to.
        sleep: Seconds to sleep after unsuccessful connections.
        timeout: Maximum number of seconds to wait for successful connections.
    """
    sleep = min(sleep, timeout) if timeout is not None else sleep
    waited = 0.0

    while True:
        try:
            start = time.perf_counter()
            with socket.create_connection((host, port), timeout=timeout):
                break
        except OSError as e:
            connection_time = time.perf_counter() - start
            waited += connection_time
            if timeout is not None and waited >= timeout:
                raise TimeoutError from e
            time.sleep(sleep)
            waited += sleep



from __future__ import annotations

import pickle

from academy.exchange.redis import RedisExchangeFactory


def test_factory_serialize(
    redis_exchange_factory: RedisExchangeFactory,
) -> None:
    pickled = pickle.dumps(redis_exchange_factory)
    reconstructed = pickle.loads(pickled)
    assert isinstance(reconstructed, RedisExchangeFactory)



from __future__ import annotations

import logging
import pickle
import uuid
from unittest import mock

import pytest

from academy.exchange import HybridExchangeFactory
from academy.exchange import HybridExchangeTransport
from academy.exchange.hybrid import base32_to_uuid
from academy.exchange.hybrid import uuid_to_base32
from academy.identifier import UserId
from academy.message import Message
from academy.message import PingRequest
from academy.socket import open_port
from testing.agents import EmptyAgent
from testing.constant import TEST_CONNECTION_TIMEOUT


@pytest.mark.asyncio
async def test_factory_serialize(
    hybrid_exchange_factory: HybridExchangeFactory,
) -> None:
    pickled = pickle.dumps(hybrid_exchange_factory)
    reconstructed = pickle.loads(pickled)
    assert isinstance(reconstructed, HybridExchangeFactory)


@pytest.mark.asyncio
async def test_key_namespaces(mock_redis) -> None:
    namespace = 'foo'
    uid = UserId.new()
    factory = HybridExchangeFactory(
        redis_host='localhost',
        redis_port=0,
        namespace=namespace,
    )
    async with await factory._create_transport() as transport:
        assert isinstance(transport, HybridExchangeTransport)

        assert transport._address_key(uid).startswith(f'{namespace}:')
        assert transport._status_key(uid).startswith(f'{namespace}:')
        assert transport._queue_key(uid).startswith(f'{namespace}:')


@pytest.mark.asyncio
async def test_send_to_mailbox_direct(
    hybrid_exchange_factory: HybridExchangeFactory,
) -> None:
    factory = hybrid_exchange_factory
    async with await factory._create_transport() as transport1:
        async with await factory._create_transport() as transport2:
            message = Message.create(
                src=transport1.mailbox_id,
                dest=transport2.mailbox_id,
                body=PingRequest(),
            )
            for _ in range(3):
                await transport1.send(message)
                response = await transport2.recv(
                    timeout=TEST_CONNECTION_TIMEOUT,
                )
                assert response == message


@pytest.mark.asyncio
async def test_send_to_mailbox_indirect(
    hybrid_exchange_factory: HybridExchangeFactory,
) -> None:
    factory = hybrid_exchange_factory
    messages = 3
    async with await factory._create_transport() as transport1:
        aid = (await transport1.register_agent(EmptyAgent)).agent_id
        message = Message.create(
            src=transport1.mailbox_id,
            dest=aid,
            body=PingRequest(),
        )
        for _ in range(messages):
            await transport1.send(message)

    async with await factory._create_transport(mailbox_id=aid) as mailbox:
        for _ in range(messages):
            received = await mailbox.recv(timeout=TEST_CONNECTION_TIMEOUT)
            assert received == message


@pytest.mark.asyncio
async def test_mailbox_redis_error_logging(
    hybrid_exchange_factory: HybridExchangeFactory,
    caplog,
) -> None:
    factory = hybrid_exchange_factory
    caplog.set_level(logging.ERROR)
    with mock.patch(
        'academy.exchange.hybrid.HybridExchangeTransport._get_message_from_redis',
        side_effect=RuntimeError('Mock thread error.'),
    ):
        async with await factory._create_transport() as transport:
            assert any(
                f'Error in redis listener task for {transport.mailbox_id}'
                in record.message
                for record in caplog.records
                if record.levelname == 'ERROR'
            )


@pytest.mark.asyncio
async def test_send_to_mailbox_bad_cached_address(
    hybrid_exchange_factory: HybridExchangeFactory,
) -> None:
    factory = hybrid_exchange_factory
    port1, port2 = open_port(), open_port()
    async with await factory._create_transport() as transport1:
        aid = (await transport1.register_agent(EmptyAgent)).agent_id

        factory1 = HybridExchangeFactory(
            redis_host='localhost',
            redis_port=0,
            ports=[port1],
        )
        async with await factory1._create_transport(
            mailbox_id=aid,
        ) as transport2:
            message = Message.create(
                src=transport1.mailbox_id,
                dest=transport2.mailbox_id,
                body=PingRequest(),
            )
            await transport1.send(message)
            received = await transport2.recv(timeout=TEST_CONNECTION_TIMEOUT)
            assert received == message

        # Address of mailbox is now in the exchanges cache but
        # the mailbox is no longer listening on that address.
        address = transport1._address_cache[transport2.mailbox_id]
        socket = transport1._socket_pool._sockets[address]
        await socket.close()

        factory2 = HybridExchangeFactory(
            redis_host='localhost',
            redis_port=0,
            ports=[port2],
        )
        async with await factory2._create_transport(
            mailbox_id=aid,
        ) as transport2:
            # This send will try the cached address, fail, catch the error,
            # and retry via redis.
            await transport1.send(message)
            received = await transport2.recv(timeout=TEST_CONNECTION_TIMEOUT)
            assert received == message


def test_uuid_encoding() -> None:
    for _ in range(3):
        uid = uuid.uuid4()
        encoded = uuid_to_base32(uid)
        assert base32_to_uuid(encoded) == uid



from __future__ import annotations

import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor

from academy.agent import action
from academy.agent import Agent
from academy.exchange import LocalExchangeFactory
from academy.logging import init_logging
from academy.manager import Manager


class Counter(Agent):
    count: int

    async def agent_on_startup(self) -> None:
        self.count = 0

    @action
    async def increment(self, value: int = 1) -> None:
        self.count += value

    @action
    async def get_count(self) -> int:
        return self.count


async def main() -> int:
    init_logging(logging.INFO)

    async with await Manager.from_exchange_factory(
        factory=LocalExchangeFactory(),
        executors=ThreadPoolExecutor(),
    ) as manager:
        agent_handle = await manager.launch(Counter)

        count = await agent_handle.get_count()
        assert count == 0

        await agent_handle.increment()

        count = await agent_handle.get_count()
        assert count == 1

    return 0


if __name__ == '__main__':
    raise SystemExit(asyncio.run(main()))



"""Create [`GlobusApp`][globus_sdk.GlobusApp] instances.

Taken from: https://github.com/proxystore/proxystore/blob/e296f937d913cae02f87ab35112fd4c8351860b0/proxystore/globus/app.py
"""

from __future__ import annotations

import os
import pathlib
from typing import Any
from typing import Literal

import click
from globus_sdk.globus_app import ClientApp
from globus_sdk.globus_app import GlobusApp
from globus_sdk.globus_app import GlobusAppConfig
from globus_sdk.globus_app import UserApp
from globus_sdk.login_flows import CommandLineLoginFlowManager
from globus_sdk.scopes import ScopeBuilder
from globus_sdk.tokenstorage import SQLiteTokenStorage
from globus_sdk.tokenstorage import TokenValidationError

# Registered `Academy-Client Application` by alokvk2@uchicago.edu
ACADEMY_GLOBUS_CLIENT_ID = '1624cf3f-45ee-4f54-9de4-2d5d79191346'

ACADEMY_GLOBUS_CLIENT_ID_ENV_NAME = 'ACADEMY_GLOBUS_CLIENT_ID'
ACADEMY_GLOBUS_CLIENT_SECRET_ENV_NAME = 'ACADEMY_GLOBUS_CLIENT_SECRET'

_APP_NAME = 'academy'
_TOKENS_FILE = 'storage.db'

AcademyExchangeScopes = ScopeBuilder(
    # "Academy Exchange Server" application client ID
    'a7e16357-8edf-414d-9e73-85e4b0b18be4',
    # The academy_exchange scope has scope ID:
    #   17619205-054c-4829-a1a8-f4b6968c76d2
    known_url_scopes=['academy_exchange'],
)


class _CustomLoginFlowManager(CommandLineLoginFlowManager):
    def print_authorize_url(
        self,
        authorize_url: str,
    ) -> None:  # pragma: no cover
        click.secho(
            'Please visit the following url to authenticate:',
            fg='cyan',
        )
        click.echo(authorize_url)

    def prompt_for_code(self) -> str:  # pragma: no cover
        auth_code = click.prompt(
            click.style('Enter the auth code:', fg='cyan'),
            prompt_suffix=' ',
        )
        return auth_code.strip()


def get_token_storage(
    filepath: str | pathlib.Path | None = None,
    *,
    namespace: str = 'DEFAULT',
) -> SQLiteTokenStorage:
    """Create token storage adapter.

    Args:
        filepath: Name of the database file. If not provided, defaults to a
            file in the Academy home directory specified by ACADEMY_HOME.
        namespace: Optional namespace to use within the database for
            partitioning token data.

    Returns:
        Token storage.
    """
    if filepath is None:
        default = os.path.join(
            os.path.expanduser('~/local/share'),
            _APP_NAME,
        )
        basepath = os.environ.get('ACADEMY_HOME', default=default)
        filepath = os.path.join(basepath, _TOKENS_FILE)

    filepath = pathlib.Path(filepath)
    filepath.parent.mkdir(parents=True, exist_ok=True)
    return SQLiteTokenStorage(filepath, namespace=namespace)


def get_client_credentials_from_env() -> tuple[str, str]:
    """Read the Globus Client ID and secret from the environment.

    The Client ID should be set to `ACADEMY_GLOBUS_CLIENT_ID` and
    the secret to `ACADEMY_GLOBUS_CLIENT_SECRET`.

    Note:
        This function performs no validation on the values of the variables.

    Returns:
        Tuple containing the client ID and secret.

    Raises:
        ValueError: if one of the environment variables is set.
    """
    try:
        client_id = os.environ[ACADEMY_GLOBUS_CLIENT_ID_ENV_NAME]
        client_secret = os.environ[ACADEMY_GLOBUS_CLIENT_SECRET_ENV_NAME]
    except KeyError as e:
        raise ValueError(
            f'Both {ACADEMY_GLOBUS_CLIENT_ID_ENV_NAME} and '
            f'{ACADEMY_GLOBUS_CLIENT_SECRET_ENV_NAME} must be set to '
            'use a client identity. Either set both environment variables '
            'or unset both to use the normal login flow.',
        ) from e

    return client_id, client_secret


def get_globus_app() -> GlobusApp:
    """Get a Globus App based on the environment.

    If a client ID and secret are set in the environment, returns a
    [`ClientApp`][globus_sdk.ClientApp] using
    [`get_client_app()`][academy.exchange.cloud.login.get_client_app].
    Otherwise returns a [`UserApp`][globus_sdk.UserApp] using
    [`get_user_app()`][academy.exchange.cloud.login.get_user_app].

    Returns:
        Initialized app.
    """
    if is_client_login():
        return get_client_app()
    return get_user_app()


def get_client_app(
    client_id: str | None = None,
    client_secret: str | None = None,
) -> ClientApp:
    """Get a Client Globus App.

    Args:
        client_id: Client ID. If one or both of the `client_id` and
            `client_secret` are not provided, the values will be read from
            the environment using
            [`get_client_credentials_from_env()`][academy.exchange.cloud.login.get_client_credentials_from_env].
        client_secret: Client secret. See above.

    Returns:
        Initialized app.
    """
    if client_id is None or client_secret is None:
        client_id, client_secret = get_client_credentials_from_env()

    config = GlobusAppConfig(
        token_storage=get_token_storage(),
        request_refresh_tokens=True,
    )

    return ClientApp(
        app_name=_APP_NAME,
        client_id=client_id,
        client_secret=client_secret,
        config=config,
    )


def get_user_app() -> UserApp:
    """Get a User Globus App.

    The [`UserApp`][globus_sdk.UserApp] will
    automatically perform an interactive flow with the user as needed.

    Returns:
        Initialized app.
    """
    config = GlobusAppConfig(
        login_flow_manager=_CustomLoginFlowManager,
        token_storage=get_token_storage(),
        request_refresh_tokens=True,
    )

    return UserApp(
        app_name=_APP_NAME,
        client_id=ACADEMY_GLOBUS_CLIENT_ID,
        config=config,
    )


def is_client_login() -> bool:
    """Check if Globus client identity environment variables are set.

    Based on the Globus Compute SDK's
    [`is_client_login()`](https://github.com/funcx-faas/funcX/blob/8f5b59075ae6f8e8b8b13fe1b91430271f4e0c3c/compute_sdk/globus_compute_sdk/sdk/login_manager/client_login.py#L24-L38){target=_blank}.

    Returns:
        `True` if `ACADEMY_GLOBUS_CLIENT_ID` and \
        `ACADEMY_GLOBUS_CLIENT_SECRET` are set.
    """
    try:
        get_client_credentials_from_env()
    except ValueError:
        return False
    else:
        return True


def get_auth_headers(
    method: Literal['globus'] | None,
    **kwargs: Any,
) -> dict[str, str]:
    """Client utility method to perform authentication and get headers."""
    if method is None:
        return {}
    elif method == 'globus':
        app = get_globus_app()
        app.add_scope_requirements(
            {
                AcademyExchangeScopes.resource_server: [
                    AcademyExchangeScopes.academy_exchange,
                ],
            },
        )
        try:
            authorizer = app.get_authorizer(
                AcademyExchangeScopes.resource_server,
            )
        except TokenValidationError:
            raise SystemExit(1) from None

        bearer = authorizer.get_authorization_header()
        assert bearer is not None
        return {'Authorization': bearer}
    else:
        raise AssertionError('Unreachable.')



from __future__ import annotations

import logging
import pathlib
import sys
import threading


class _Formatter(logging.Formatter):
    def __init__(self, color: bool = False, extra: bool = False) -> None:
        if color:
            grey = '\033[2;37m'
            green = '\033[32m'
            cyan = '\033[36m'
            blue = '\033[34m'
            yellow = '\033[33m'
            red = '\033[31m'
            purple = '\033[35m'
            reset = '\033[0m'
        else:
            grey = ''
            green = ''
            cyan = ''
            blue = ''
            yellow = ''
            red = ''
            purple = ''
            reset = ''

        if extra:
            extra_fmt = f'{green}[tid=%(os_thread)d pid=%(process)d]{reset} '
        else:
            extra_fmt = ''

        datefmt = '%Y-%m-%d %H:%M:%S'
        logfmt = (
            f'{grey}[%(asctime)s.%(msecs)03d]{reset} {extra_fmt}'
            f'{{level}}%(levelname)-8s{reset} '
            f'{purple}(%(name)s){reset} %(message)s'
        )
        debug_fmt = logfmt.format(level=cyan)
        info_fmt = logfmt.format(level=blue)
        warning_fmt = logfmt.format(level=yellow)
        error_fmt = logfmt.format(level=red)

        self.formatters = {
            logging.DEBUG: logging.Formatter(debug_fmt, datefmt=datefmt),
            logging.INFO: logging.Formatter(info_fmt, datefmt=datefmt),
            logging.WARNING: logging.Formatter(warning_fmt, datefmt=datefmt),
            logging.ERROR: logging.Formatter(error_fmt, datefmt=datefmt),
            logging.CRITICAL: logging.Formatter(error_fmt, datefmt=datefmt),
        }

    def format(self, record: logging.LogRecord) -> str:  # pragma: no cover
        return self.formatters[record.levelno].format(record)


def _os_thread_filter(
    record: logging.LogRecord,
) -> logging.LogRecord:  # pragma: no cover
    record.os_thread = threading.get_native_id()
    return record


def init_logging(  # noqa: PLR0913
    level: int | str = logging.INFO,
    *,
    logfile: str | pathlib.Path | None = None,
    logfile_level: int | str | None = None,
    color: bool = True,
    extra: bool = False,
    force: bool = False,
) -> logging.Logger:
    """Initialize global logger.

    Args:
        level: Minimum logging level.
        logfile: Configure a file handler for this path.
        logfile_level: Minimum logging level for the file handler. Defaults
            to that of `level`.
        color: Use colorful logging for stdout.
        extra: Include extra info in log messages, such as thread ID and
            process ID. This is helpful for debugging.
        force: Remove any existing handlers attached to the root
            handler. This option is useful to silencing the third-party
            package logging. Note: should not be set when running inside
            pytest.

    Returns:
        The root logger.
    """
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(_Formatter(color=color, extra=extra))
    stdout_handler.setLevel(level)
    if extra:
        stdout_handler.addFilter(_os_thread_filter)
    handlers: list[logging.Handler] = [stdout_handler]

    if logfile is not None:
        logfile_level = level if logfile_level is None else logfile_level
        path = pathlib.Path(logfile)
        path.parent.mkdir(parents=True, exist_ok=True)
        handler = logging.FileHandler(path)
        handler.setFormatter(_Formatter(color=False, extra=extra))
        handler.setLevel(logfile_level)
        handlers.append(handler)

    logging.basicConfig(
        datefmt='%Y-%m-%d %H:%M:%S',
        level=logging.NOTSET,
        handlers=handlers,
        force=force,
    )

    # This needs to be after the configuration of the root logger because
    # warnings get logged to a 'py.warnings' logger.
    logging.captureWarnings(True)

    logger = logging.getLogger()
    logger.info(
        'Configured logger (stdout-level=%s, logfile=%s, logfile-level=%s)',
        logging.getLevelName(level) if isinstance(level, int) else level,
        logfile,
        logging.getLevelName(logfile_level)
        if isinstance(logfile_level, int)
        else logfile_level,
    )

    return logger



from __future__ import annotations

import asyncio
import random
import socket
import string
import sys
from collections.abc import AsyncGenerator
from unittest import mock

import pytest
import pytest_asyncio

from academy.socket import _BAD_FILE_DESCRIPTOR_ERRNO
from academy.socket import _make_header
from academy.socket import address_by_hostname
from academy.socket import address_by_interface
from academy.socket import MESSAGE_CHUNK_SIZE
from academy.socket import open_port
from academy.socket import SimpleSocket
from academy.socket import SimpleSocketServer
from academy.socket import SocketClosedError
from academy.socket import SocketOpenError
from academy.socket import SocketPool
from academy.socket import wait_connection


@pytest_asyncio.fixture(loop_scope='module')
async def echo_server() -> AsyncGenerator[tuple[str, int]]:
    async def _handler(
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        data = await reader.read(100)
        writer.write(data)
        await writer.drain()
        writer.close()
        await writer.wait_closed()

    host, port = 'localhost', open_port()
    server = await asyncio.start_server(_handler, host, port)

    async with server:
        await server.start_serving()
        yield host, port


@pytest.mark.parametrize('shutdown', (True, False))
@pytest.mark.asyncio(loop_scope='module')
async def test_create_simple_socket(
    echo_server: tuple[str, int],
    shutdown: bool,
) -> None:
    host, port = echo_server
    async with await SimpleSocket.connect(host, port) as socket:
        assert isinstance(repr(socket), str)
        assert isinstance(str(socket), str)

        await socket.close(shutdown=shutdown)
        with pytest.raises(SocketClosedError):
            await socket.recv()


@pytest.mark.asyncio
async def test_create_simple_socket_error() -> None:
    with mock.patch('asyncio.open_connection', side_effect=OSError()):
        with pytest.raises(SocketOpenError):
            await SimpleSocket.connect('localhost', 0)


@pytest.mark.asyncio(loop_scope='module')
async def test_simple_socket_send_recv(echo_server: tuple[str, int]) -> None:
    host, port = echo_server
    async with await SimpleSocket.connect(host, port) as socket:
        # Should be a no-op
        await socket.send(b'')

        message = 'hello, world!'
        await socket.send_string(message)
        assert await socket.recv_string() == message

        error = OSError(_BAD_FILE_DESCRIPTOR_ERRNO, 'Bad file descriptor.')
        with mock.patch.object(socket.writer, 'write', side_effect=error):
            with pytest.raises(OSError, match='Bad file descriptor.'):
                await socket.send_string('hello, again!')


@pytest.mark.asyncio(loop_scope='module')
async def test_simple_socket_send_multipart(
    echo_server: tuple[str, int],
) -> None:
    host, port = echo_server
    size = int(2.5 * MESSAGE_CHUNK_SIZE)
    message = ''.join(
        [random.choice(string.ascii_uppercase) for _ in range(size)],
    )
    async with await SimpleSocket.connect(host, port) as socket:
        with mock.patch.object(socket.writer, 'write') as mock_write:
            await socket.send_string(message)
            assert mock_write.call_count == 4  # noqa: PLR2004


@pytest.mark.asyncio(loop_scope='module')
async def test_simple_socket_recv_multipart(
    echo_server: tuple[str, int],
) -> None:
    host, port = echo_server
    async with await SimpleSocket.connect(host, port) as socket:
        messages = [b'hello, world!', b'part2']
        with mock.patch.object(socket.reader, 'readexactly') as mock_read:
            # Mock received parts to include two messages split across three
            # parts followed by a socket close event.
            mock_read.side_effect = [
                _make_header(messages[0]),
                b'hello, ',
                b'world!',
                _make_header(messages[1]),
                messages[1],
                SocketClosedError(),
            ]

            assert await socket.recv_string() == 'hello, world!'
            assert await socket.recv_string() == 'part2'
            with pytest.raises(SocketClosedError):
                assert await socket.recv_string()


@pytest.mark.asyncio(loop_scope='module')
async def test_socket_pool_send_and_reuse_socket(
    echo_server: tuple[str, int],
) -> None:
    host, port = echo_server
    address = f'{host}:{port}'

    pool = SocketPool()
    message = b'hello'

    await pool.send(address, message)
    await pool.send(address, message)

    assert address in pool._sockets
    assert len(pool._sockets) == 1

    await pool.close()


@pytest.mark.asyncio(loop_scope='module')
async def test_socket_pool_close_removes_socket(echo_server: tuple[str, int]):
    host, port = echo_server
    address = f'{host}:{port}'

    pool = SocketPool()
    await pool.send(address, b'ping')

    assert address in pool._sockets

    await pool.close_socket(address)
    assert address not in pool._sockets


@pytest.mark.asyncio(loop_scope='module')
async def test_pool_close_closes_all(echo_server: tuple[str, int]):
    host, port = echo_server
    address = f'{host}:{port}'

    pool = SocketPool()
    await pool.send(address, b'ping')
    await pool.close()

    assert len(pool._sockets) == 0


@pytest_asyncio.fixture(loop_scope='module')
async def simple_socket_server() -> AsyncGenerator[SimpleSocketServer]:
    async def _identity(x: bytes) -> bytes:
        return x

    server = SimpleSocketServer(handler=_identity, host='localhost', port=None)
    async with server.serve():
        yield server


@pytest.mark.asyncio(loop_scope='module')
async def test_simple_socket_server_connect(
    simple_socket_server: SimpleSocketServer,
) -> None:
    for _ in range(3):
        async with await SimpleSocket.connect(
            simple_socket_server.host,
            simple_socket_server.port,
        ):
            pass


@pytest.mark.asyncio(loop_scope='module')
async def test_simple_socket_server_ping_pong(
    simple_socket_server: SimpleSocketServer,
) -> None:
    message = 'hello, world!'
    async with await SimpleSocket.connect(
        simple_socket_server.host,
        simple_socket_server.port,
    ) as socket:
        for _ in range(3):
            await socket.send_string(message)
            assert await socket.recv_string() == message


@pytest.mark.asyncio(loop_scope='module')
async def test_simple_socket_server_packed(
    simple_socket_server: SimpleSocketServer,
) -> None:
    # Pack many messages into one buffer to be send
    messages = [b'first message', b'seconds message', b'third message']
    buffer = b''.join(_make_header(m) + m for m in messages)

    async with await SimpleSocket.connect(
        simple_socket_server.host,
        simple_socket_server.port,
    ) as socket:
        socket.writer.write(buffer)
        await socket.writer.drain()
        for expected in messages:
            assert await socket.recv() == expected


@pytest.mark.asyncio(loop_scope='module')
async def test_simple_socket_server_multipart(
    simple_socket_server: SimpleSocketServer,
) -> None:
    # Generate >1024 bytes of data since _recv_from_socket reads in 1kB
    # chunks. This test forces recv() to buffer incomplete chunks.
    first_parts = [random.randbytes(500) for _ in range(3)]
    second_part = b'second message!'
    # socket.recv_string() will not return the delimiter so add after
    # computing the expected string
    first_expected = b''.join(first_parts)
    parts = [
        _make_header(first_expected),
        first_parts[0],
        first_parts[1],
        first_parts[2],
        _make_header(second_part),
        second_part,
    ]

    async with await SimpleSocket.connect(
        simple_socket_server.host,
        simple_socket_server.port,
    ) as socket:
        for part in parts:
            socket.writer.write(part)
        await socket.writer.drain()
        assert await socket.recv() == first_expected
        assert await socket.recv() == second_part


@pytest.mark.asyncio(loop_scope='module')
async def test_simple_socket_server_client_disconnect_early(
    simple_socket_server: SimpleSocketServer,
) -> None:
    async with await SimpleSocket.connect(
        simple_socket_server.host,
        simple_socket_server.port,
    ):
        # Client disconnects without sending anything
        pass


def test_get_address_by_hostname() -> None:
    assert isinstance(address_by_hostname(), str)


@pytest.mark.skipif(
    sys.platform == 'darwin',
    reason='Test does not run on darwin',
)
def test_get_address_by_interface() -> None:  # pragma: darwin no cover
    for _, ifname in socket.if_nameindex():
        try:
            assert isinstance(address_by_interface(ifname), str)
        except Exception:  # pragma: no cover
            continue
        else:
            break
    else:  # pragma: no cover
        raise RuntimeError('Failed to find a valid address by interface.')


def test_wait_connection() -> None:
    with mock.patch('socket.create_connection'):
        wait_connection('localhost', port=0)


def test_wait_connection_timeout() -> None:
    with mock.patch('socket.create_connection', side_effect=OSError()):
        with pytest.raises(TimeoutError):
            wait_connection('localhost', port=0, sleep=0, timeout=0)

        with pytest.raises(TimeoutError):
            wait_connection('localhost', port=0, sleep=0.01, timeout=0.05)



from __future__ import annotations

import logging
import pickle
from unittest import mock

import aiohttp
import pytest

from academy.exception import BadEntityIdError
from academy.exception import ForbiddenError
from academy.exception import MailboxTerminatedError
from academy.exception import UnauthorizedError
from academy.exchange import HttpExchangeFactory
from academy.exchange import HttpExchangeTransport
from academy.exchange.cloud.client import _raise_for_status
from academy.exchange.cloud.client import spawn_http_exchange
from academy.exchange.cloud.server import StatusCode
from academy.identifier import UserId
from academy.socket import open_port
from testing.constant import TEST_CONNECTION_TIMEOUT


def test_factory_serialize(
    http_exchange_factory: HttpExchangeFactory,
) -> None:
    pickled = pickle.dumps(http_exchange_factory)
    reconstructed = pickle.loads(pickled)
    assert isinstance(reconstructed, HttpExchangeFactory)


@pytest.mark.asyncio
async def test_recv_timeout(http_exchange_server: tuple[str, int]) -> None:
    host, port = http_exchange_server
    url = f'http://{host}:{port}'
    factory = HttpExchangeFactory(url)

    class MockResponse:
        status = StatusCode.TIMEOUT.value

        async def __aexit__(self, exc_type, exc, tb):
            pass

        async def __aenter__(self):
            return self

    async with await factory._create_transport() as transport:
        response = MockResponse()
        with mock.patch.object(
            transport._session,
            'get',
            return_value=response,
        ):
            with pytest.raises(TimeoutError):
                await transport.recv()


@pytest.mark.asyncio
async def test_additional_headers(
    http_exchange_server: tuple[str, int],
) -> None:
    host, port = http_exchange_server
    url = f'http://{host}:{port}'
    headers = {'Authorization': 'fake auth'}
    factory = HttpExchangeFactory(url, additional_headers=headers)
    async with await factory._create_transport() as transport:
        assert isinstance(transport, HttpExchangeTransport)
        assert 'Authorization' in transport._session.headers


async def test_raise_for_status_error_conversion() -> None:
    class _MockResponse(aiohttp.ClientResponse):
        def __init__(self, status: int) -> None:
            self.status = status

    response = _MockResponse(StatusCode.OKAY.value)
    _raise_for_status(response, UserId.new())

    response = _MockResponse(StatusCode.UNAUTHORIZED.value)
    with pytest.raises(UnauthorizedError):
        _raise_for_status(response, UserId.new())

    response = _MockResponse(StatusCode.FORBIDDEN.value)
    with pytest.raises(ForbiddenError):
        _raise_for_status(response, UserId.new())

    response = _MockResponse(StatusCode.NOT_FOUND.value)
    with pytest.raises(BadEntityIdError):
        _raise_for_status(response, UserId.new())

    response = _MockResponse(StatusCode.TERMINATED.value)
    with pytest.raises(MailboxTerminatedError):
        _raise_for_status(response, UserId.new())

    response = _MockResponse(StatusCode.TIMEOUT.value)
    with pytest.raises(TimeoutError):
        _raise_for_status(response, UserId.new())


@pytest.mark.asyncio
async def test_spawn_http_exchange() -> None:
    with spawn_http_exchange(
        'localhost',
        open_port(),
        level=logging.ERROR,
        timeout=TEST_CONNECTION_TIMEOUT,
    ) as factory:
        async with await factory._create_transport() as transport:
            assert isinstance(transport, HttpExchangeTransport)



from __future__ import annotations

import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor

from academy.agent import action
from academy.agent import Agent
from academy.exchange import LocalExchangeFactory
from academy.handle import Handle
from academy.logging import init_logging
from academy.manager import Manager

logger = logging.getLogger(__name__)


class Coordinator(Agent):
    def __init__(
        self,
        lowerer: Handle[Lowerer],
        reverser: Handle[Reverser],
    ) -> None:
        super().__init__()
        self.lowerer = lowerer
        self.reverser = reverser

    @action
    async def process(self, text: str) -> str:
        text = await self.lowerer.lower(text)
        text = await self.reverser.reverse(text)
        return text


class Lowerer(Agent):
    @action
    async def lower(self, text: str) -> str:
        return text.lower()


class Reverser(Agent):
    @action
    async def reverse(self, text: str) -> str:
        return text[::-1]


async def main() -> int:
    init_logging(logging.INFO)

    async with await Manager.from_exchange_factory(
        factory=LocalExchangeFactory(),
        executors=ThreadPoolExecutor(),
    ) as manager:
        lowerer = await manager.launch(Lowerer)
        reverser = await manager.launch(Reverser)
        coordinator = await manager.launch(
            Coordinator,
            args=(lowerer, reverser),
        )

        text = 'DEADBEEF'
        expected = 'feebdaed'

        logger.info('Invoking process("%s") on %s', text, coordinator.agent_id)
        result = await coordinator.process(text)
        assert result == expected
        logger.info('Received result: "%s"', result)

    return 0


if __name__ == '__main__':
    raise SystemExit(asyncio.run(main()))



# ruff: noqa: D102
from __future__ import annotations

import asyncio
import contextlib
import dataclasses
import logging
import multiprocessing
import sys
import uuid
from collections.abc import Generator
from typing import Any
from typing import Generic
from typing import Literal
from typing import NamedTuple
from urllib.parse import urlparse

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    from typing import Self
else:  # pragma: <3.11 cover
    from typing_extensions import Self

import aiohttp

from academy.agent import Agent
from academy.agent import AgentT
from academy.exception import BadEntityIdError
from academy.exception import ForbiddenError
from academy.exception import MailboxTerminatedError
from academy.exception import UnauthorizedError
from academy.exchange.cloud.config import ExchangeServingConfig
from academy.exchange.cloud.login import get_auth_headers
from academy.exchange.cloud.server import _run
from academy.exchange.cloud.server import StatusCode
from academy.exchange.factory import ExchangeFactory
from academy.exchange.transport import ExchangeTransportMixin
from academy.exchange.transport import MailboxStatus
from academy.identifier import AgentId
from academy.identifier import EntityId
from academy.identifier import UserId
from academy.message import Message
from academy.serialize import NoPickleMixin
from academy.socket import wait_connection

logger = logging.getLogger(__name__)


class _HttpConnectionInfo(NamedTuple):
    url: str
    additional_headers: dict[str, str] | None = None
    ssl_verify: bool | None = None


@dataclasses.dataclass
class HttpAgentRegistration(Generic[AgentT]):
    """Agent registration for Http exchanges."""

    agent_id: AgentId[AgentT]
    """Unique identifier for the agent created by the exchange."""


class HttpExchangeTransport(ExchangeTransportMixin, NoPickleMixin):
    """Http exchange client.

    Args:
        mailbox_id: Identifier of the mailbox on the exchange. If there is
            not an id provided, the exchange will create a new client mailbox.
        session: Http session.
        connection_info: Exchange connection info.
    """

    def __init__(
        self,
        mailbox_id: EntityId,
        session: aiohttp.ClientSession,
        connection_info: _HttpConnectionInfo,
    ) -> None:
        self._mailbox_id = mailbox_id
        self._session = session
        self._info = connection_info

        base_url = self._info.url
        self._mailbox_url = f'{base_url}/mailbox'
        self._message_url = f'{base_url}/message'
        self._discover_url = f'{base_url}/discover'

    @classmethod
    async def new(
        cls,
        *,
        connection_info: _HttpConnectionInfo,
        mailbox_id: EntityId | None = None,
        name: str | None = None,
    ) -> Self:
        """Instantiate a new transport.

        Args:
            connection_info: Exchange connection information.
            mailbox_id: Bind the transport to the specific mailbox. If `None`,
                a new user entity will be registered and the transport will be
                bound to that mailbox.
            name: Display name of the registered entity if `mailbox_id` is
                `None`.

        Returns:
            An instantiated transport bound to a specific mailbox.
        """
        ssl_verify = connection_info.ssl_verify
        if ssl_verify is None:  # pragma: no branch
            scheme = urlparse(connection_info.url).scheme
            ssl_verify = scheme == 'https'

        session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=ssl_verify),
            headers=connection_info.additional_headers,
        )

        if mailbox_id is None:
            mailbox_id = UserId.new(name=name)
            async with session.post(
                f'{connection_info.url}/mailbox',
                json={'mailbox': mailbox_id.model_dump_json()},
            ) as response:
                _raise_for_status(response, mailbox_id)
            logger.info('Registered %s in exchange', mailbox_id)

        return cls(mailbox_id, session, connection_info)

    @property
    def mailbox_id(self) -> EntityId:
        return self._mailbox_id

    async def close(self) -> None:
        await self._session.close()

    async def discover(
        self,
        agent: type[Agent],
        *,
        allow_subclasses: bool = True,
    ) -> tuple[AgentId[Any], ...]:
        agent_str = f'{agent.__module__}.{agent.__name__}'
        async with self._session.get(
            self._discover_url,
            json={
                'agent': agent_str,
                'allow_subclasses': allow_subclasses,
            },
        ) as response:
            _raise_for_status(response, self.mailbox_id)
            agent_ids_str = (await response.json())['agent_ids']
        agent_ids = [aid for aid in agent_ids_str.split(',') if len(aid) > 0]
        return tuple(AgentId(uid=uuid.UUID(aid)) for aid in agent_ids)

    def factory(self) -> HttpExchangeFactory:
        # Note: When getting factory, auth method is not preserved
        # but auth headers (i.e. the bearer token) is.
        return HttpExchangeFactory(
            url=self._info.url,
            additional_headers=self._info.additional_headers,
            ssl_verify=self._info.ssl_verify,
        )

    async def recv(self, timeout: float | None = None) -> Message[Any]:
        try:
            async with self._session.get(
                self._message_url,
                json={
                    'mailbox': self.mailbox_id.model_dump_json(),
                    'timeout': timeout,
                },
                timeout=aiohttp.ClientTimeout(timeout),
            ) as response:
                _raise_for_status(response, self.mailbox_id)
                message_raw = (await response.json()).get('message')
        except asyncio.TimeoutError as e:
            # In older versions of Python, ayncio.TimeoutError and TimeoutError
            # are different types.
            raise TimeoutError(
                f'Failed to receive response in {timeout} seconds.',
            ) from e

        return Message.model_validate_json(message_raw)

    async def register_agent(
        self,
        agent: type[AgentT],
        *,
        name: str | None = None,
    ) -> HttpAgentRegistration[AgentT]:
        aid: AgentId[AgentT] = AgentId.new(name=name)
        async with self._session.post(
            self._mailbox_url,
            json={
                'mailbox': aid.model_dump_json(),
                'agent': ','.join(agent._agent_mro()),
            },
        ) as response:
            _raise_for_status(response, self.mailbox_id, aid)
        return HttpAgentRegistration(agent_id=aid)

    async def send(self, message: Message[Any]) -> None:
        async with self._session.put(
            self._message_url,
            json={'message': message.model_dump_json()},
        ) as response:
            _raise_for_status(response, self.mailbox_id, message.dest)

    async def status(self, uid: EntityId) -> MailboxStatus:
        async with self._session.get(
            self._mailbox_url,
            json={'mailbox': uid.model_dump_json()},
        ) as response:
            _raise_for_status(response, self.mailbox_id, uid)
            status = (await response.json())['status']
            return MailboxStatus(status)

    async def terminate(self, uid: EntityId) -> None:
        async with self._session.delete(
            self._mailbox_url,
            json={'mailbox': uid.model_dump_json()},
        ) as response:
            _raise_for_status(response, self.mailbox_id, uid)


class HttpExchangeFactory(ExchangeFactory[HttpExchangeTransport]):
    """Http exchange client factory.

    Args:
        url: Address of HTTP exchange
        auth_method: Method to get authorization headers
        additional_headers: Any other information necessary to communicate
            with the exchange. Used for passing the Globus bearer token
        ssl_verify: Same as requests.Session.verify. If the server's TLS
            certificate should be validated. Should be true if using HTTPS
            Only set to false for testing or local development.
    """

    def __init__(
        self,
        url: str,
        auth_method: Literal['globus'] | None = None,
        additional_headers: dict[str, str] | None = None,
        ssl_verify: bool | None = None,
    ) -> None:
        if additional_headers is None:
            additional_headers = {}
        additional_headers |= get_auth_headers(auth_method)

        self._info = _HttpConnectionInfo(
            url=url,
            additional_headers=additional_headers,
            ssl_verify=ssl_verify,
        )

    async def _create_transport(
        self,
        mailbox_id: EntityId | None = None,
        *,
        name: str | None = None,
        registration: HttpAgentRegistration[Any] | None = None,  # type: ignore[override]
    ) -> HttpExchangeTransport:
        return await HttpExchangeTransport.new(
            connection_info=self._info,
            mailbox_id=mailbox_id,
            name=name,
        )


def _raise_for_status(
    response: aiohttp.ClientResponse,
    client_id: EntityId,
    resource_id: EntityId | None = None,
) -> None:
    # Parse HTTP error codes into the correct error types.
    #   - client_id is the ID of the transport client that is making
    #     the request.
    #   - resource_id is the ID of the resource being accessed in the case
    #     of operations like send/status/terminate.
    if response.status == StatusCode.UNAUTHORIZED.value:
        raise UnauthorizedError(
            f'Exchange entity {client_id} does not have the required '
            'authorization credentials.',
        )
    elif response.status == StatusCode.FORBIDDEN.value:
        raise ForbiddenError(
            f'Exchange entity {client_id} is not authorized to access '
            'this resource.',
        )
    elif response.status == StatusCode.NOT_FOUND.value:
        raise BadEntityIdError(resource_id or client_id)
    elif response.status == StatusCode.TERMINATED.value:
        raise MailboxTerminatedError(resource_id or client_id)
    elif response.status == StatusCode.TIMEOUT.value:
        raise TimeoutError()
    else:
        response.raise_for_status()


@contextlib.contextmanager
def spawn_http_exchange(
    host: str = '0.0.0.0',
    port: int = 5463,
    *,
    level: int | str = logging.WARNING,
    timeout: float | None = None,
) -> Generator[HttpExchangeFactory]:
    """Context manager that spawns an HTTP exchange in a subprocess.

    This function spawns a new process (rather than forking) and wait to
    return until a connection with the exchange has been established.
    When exiting the context manager, `SIGINT` will be sent to the exchange
    process. If the process does not exit within 5 seconds, it will be
    killed.

    Warning:
        The exclusion of authentication and ssl configuration is
        intentional. This method should only be used for temporary exchanges
        in trusted environments (i.e. the login node of a cluster).

    Args:
        host: Host the exchange should listen on.
        port: Port the exchange should listen on.
        level: Logging level.
        timeout: Connection timeout when waiting for exchange to start.

    Returns:
        Exchange interface connected to the spawned exchange.
    """
    # Fork is not safe in multi-threaded context.
    multiprocessing.set_start_method('spawn')

    config = ExchangeServingConfig(host=host, port=port, log_level=level)
    exchange_process = multiprocessing.Process(
        target=_run,
        args=(config,),
    )
    exchange_process.start()

    logger.info('Starting exchange server...')
    wait_connection(host, port, timeout=timeout)
    logger.info('Started exchange server!')

    base_url = f'http://{host}:{port}'
    factory = HttpExchangeFactory(base_url)
    try:
        yield factory
    finally:
        logger.info('Terminating exchange server...')
        wait = 5
        exchange_process.terminate()
        exchange_process.join(timeout=wait)
        if exchange_process.exitcode is None:  # pragma: no cover
            logger.info(
                'Killing exchange server after waiting %s seconds',
                wait,
            )
            exchange_process.kill()
        else:
            logger.info('Terminated exchange server!')
        exchange_process.close()



"""Authenticate users from request headers."""

from __future__ import annotations

import uuid
from collections.abc import Mapping
from threading import Lock
from typing import Protocol
from typing import runtime_checkable

import globus_sdk

from academy.exchange.cloud.config import ExchangeAuthConfig
from academy.exchange.cloud.exceptions import ForbiddenError
from academy.exchange.cloud.exceptions import UnauthorizedError
from academy.exchange.cloud.login import AcademyExchangeScopes


@runtime_checkable
class Authenticator(Protocol):
    """Authenticate users from request headers."""

    def authenticate_user(self, headers: Mapping[str, str]) -> uuid.UUID:
        """Authenticate user from request headers.

        Warning:
            This method must be thread safe!

        Args:
            headers: Request headers.

        Returns:
            A user id upon authentication success.

        Raises:
            ForbiddenError: user is authenticated but is missing permissions
                or accessing forbidden resources.
            UnauthorizedError: user authentication fails.
        """
        ...


class NullAuthenticator:
    """Authenticator that implements no authentication."""

    def authenticate_user(self, headers: Mapping[str, str]) -> uuid.UUID:
        """Authenticate user from request headers.

        Args:
            headers: Request headers.

        Returns:
            Null user regardless of provided headers.
        """
        return uuid.UUID(int=0)


class GlobusAuthenticator:
    """Globus Auth authorizer.

    Args:
        client_id: Globus application client ID. If either `client_id`
            or `client_secret` is `None`, the values will be read from the
            environment variables as described in
            [`get_confidential_app_auth_client`][proxystore.globus.client.get_confidential_app_auth_client].
            Ignored if `auth_client` is provided.
        client_secret: Globus application client secret. See `client_id` for
            details. Ignored if `auth_client` is provided.
        audience: Intended audience of the token. This should typically be
            the resource server of the the token was issued for. E.g.,
            the UUID of the ProxyStore Relay Server application.
        auth_client: Optional confidential application authentication client
            which is used for introspecting client tokens.
    """

    def __init__(
        self,
        client_id: str | None = None,
        client_secret: str | None = None,
        *,
        audience: str = AcademyExchangeScopes.resource_server,
        auth_client: globus_sdk.ConfidentialAppAuthClient | None = None,
    ) -> None:
        self.auth_client = (
            globus_sdk.ConfidentialAppAuthClient(
                client_id=str(client_id),
                client_secret=str(client_secret),
            )
            if auth_client is None
            else auth_client
        )
        self.auth_client_lock = Lock()
        self.audience = audience

    def authenticate_user(self, headers: Mapping[str, str]) -> uuid.UUID:
        """Authenticate a Globus Auth user from request header.

        This follows from the [Globus Sample Data Portal](https://github.com/globus/globus-sample-data-portal/blob/30e30cd418ee9b103e04916e19deb9902d3aafd8/service/decorators.py)
        example.

        The underlying auth client is not thread safe, but this method is made
        thread safe using a lock.

        Args:
            headers: Request headers to extract tokens from.

        Returns:
            Globus Auth identity returned via \
            [token introspection](https://docs.globus.org/api/auth/reference/#token-introspect).

        Raises:
            UnauthorizedError: if the authorization header is missing or
                the header is malformed.
            ForbiddenError: if the tokens have expired or been revoked.
            ForbiddenError: if `audience` is not included in the token's
                audience.
        """
        token = get_token_from_headers(headers)
        with self.auth_client_lock:
            token_meta = self.auth_client.oauth2_token_introspect(token)

        if not token_meta.get('active'):
            raise ForbiddenError('Token is expired or has been revoked.')

        if self.audience is not None and self.audience not in token_meta.get(
            'aud',
            [],
        ):
            raise ForbiddenError(
                f'Token audience does not include "{self.audience}". This '
                'could result in a confused deputy attack. Ensure the correct '
                'scopes are requested when the token is created.',
            )

        return uuid.UUID(token_meta.get('client_id'))


def get_authenticator(config: ExchangeAuthConfig) -> Authenticator:
    """Create an authenticator from a configuration.

    Args:
        config: Configuration.

    Returns:
        Authenticator.

    Raises:
        ValueError: if the authentication method in the config is unknown.
    """
    if config.method is None:
        return NullAuthenticator()
    elif config.method == 'globus':
        return GlobusAuthenticator(**config.kwargs)
    else:
        raise ValueError(f'Unknown authentication method "{config.method}."')


def get_token_from_headers(headers: Mapping[str, str]) -> str:
    """Extract token from websockets headers.

    The header is expected to have the format `Authorization: Bearer <TOKEN>`.

    Args:
         headers: Request headers to extract tokens from.

    Returns:
        String token.

    Raises:
        UnauthorizedError: if the authorization header is missing.
        UnauthorizedError: if the authorization header is malformed.
    """
    if 'Authorization' not in headers:
        raise UnauthorizedError(
            'Request headers are missing authorization header.',
        )

    auth_header_parts = headers['Authorization'].split(' ')

    if len(auth_header_parts) != 2 or auth_header_parts[0] != 'Bearer':  # noqa: PLR2004
        raise UnauthorizedError(
            'Bearer token in authorization header is malformed.',
        )

    return auth_header_parts[1]
